#!/usr/bin/env python
"""
_ParentlessMergeBySize_t_

Unit tests for parentless WMBS merging.
"""

import unittest
import threading

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.DataStructs.Run import Run
from WMCore.DAOFactory import DAOFactory
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMQuality.TestInit import TestInit


class ParentlessMergeBySizeTest(unittest.TestCase):
    """
    _ParentlessMergeBySizeTest_

    Unit tests for parentless WMBS merging.
    """

    def setUp(self):
        """
        _setUp_

        Boiler plate DB setup.
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        return

    def tearDown(self):
        """
        _tearDown_

        Clear out WMBS.
        """
        self.testInit.clearDatabase()
        return

    def stuffWMBS(self):
        """
        _stuffWMBS_

        Insert some dummy jobs, jobgroups, filesets, files and subscriptions
        into WMBS to test job creation.  Three completed job groups each
        containing several files are injected.  Another incomplete job group is
        also injected.  Also files are added to the "Mergeable" subscription as
        well as to the output fileset for their jobgroups.
        """
        locationAction = self.daoFactory(classname="Locations.New")
        locationAction.execute(siteName="T1_US_FNAL", pnn="T1_US_FNAL_Disk")

        self.mergeFileset = Fileset(name="mergeFileset")
        self.mergeFileset.create()
        self.bogusFileset = Fileset(name="bogusFileset")
        self.bogusFileset.create()

        mergeWorkflow = Workflow(name="mergeWorkflow", spec="bunk2",
                                 owner="Steve", task="Test")
        mergeWorkflow.create()
        markWorkflow = self.daoFactory(classname="Workflow.MarkInjectedWorkflows")
        markWorkflow.execute(names=[mergeWorkflow.name], injected=True)

        self.mergeSubscription = Subscription(fileset=self.mergeFileset,
                                              workflow=mergeWorkflow,
                                              split_algo="ParentlessMergeBySize")
        self.mergeSubscription.create()
        self.bogusSubscription = Subscription(fileset=self.bogusFileset,
                                              workflow=mergeWorkflow,
                                              split_algo="ParentlessMergeBySize")

        file1 = File(lfn = "file1", size = 1024, events = 1024, first_event = 0,
                     locations = set(["T1_US_FNAL_Disk"]))
        file1.addRun(Run(1, *[45]))
        file1.create()
        file2 = File(lfn = "file2", size = 1024, events = 1024,
                     first_event = 1024, locations = set(["T1_US_FNAL_Disk"]))
        file2.addRun(Run(1, *[45]))
        file2.create()
        file3 = File(lfn = "file3", size = 1024, events = 1024,
                     first_event = 2048, locations = set(["T1_US_FNAL_Disk"]))
        file3.addRun(Run(1, *[45]))
        file3.create()
        file4 = File(lfn = "file4", size = 1024, events = 1024,
                     first_event = 3072, locations = set(["T1_US_FNAL_Disk"]))
        file4.addRun(Run(1, *[45]))
        file4.create()

        fileA = File(lfn = "fileA", size = 1024, events = 1024,
                     first_event = 0, locations = set(["T1_US_FNAL_Disk"]))
        fileA.addRun(Run(1, *[46]))
        fileA.create()
        fileB = File(lfn = "fileB", size = 1024, events = 1024,
                     first_event = 1024, locations = set(["T1_US_FNAL_Disk"]))
        fileB.addRun(Run(1, *[46]))
        fileB.create()
        fileC = File(lfn = "fileC", size = 1024, events = 1024,
                     first_event = 2048, locations = set(["T1_US_FNAL_Disk"]))
        fileC.addRun(Run(1, *[46]))
        fileC.create()

        fileI = File(lfn = "fileI", size = 1024, events = 1024,
                     first_event = 0, locations = set(["T1_US_FNAL_Disk"]))
        fileI.addRun(Run(2, *[46]))
        fileI.create()
        fileII = File(lfn = "fileII", size = 1024, events = 1024,
                      first_event = 1024, locations = set(["T1_US_FNAL_Disk"]))
        fileII.addRun(Run(2, *[46]))
        fileII.create()
        fileIII = File(lfn = "fileIII", size = 1024, events = 102400,
                       first_event = 2048, locations = set(["T1_US_FNAL_Disk"]))
        fileIII.addRun(Run(2, *[46]))
        fileIII.create()
        fileIV = File(lfn = "fileIV", size = 102400, events = 1024,
                      first_event = 3072, locations = set(["T1_US_FNAL_Disk"]))
        fileIV.addRun(Run(2, *[46]))
        fileIV.create()

        for jobFile in [file1, file2, file3, file4, fileA, fileB, fileC, fileI,
                        fileII, fileIII, fileIV]:
            self.mergeFileset.addFile(jobFile)
            self.bogusFileset.addFile(jobFile)

        self.mergeFileset.commit()
        self.bogusFileset.commit()

        return

    def testMinMergeSize1(self):
        """
        _testMinMergeSize1_

        Set the minimum merge size to be 20,000 bytes which is more than the
        sum of all file sizes in the WMBS instance.  Verify that no merge jobs
        will be produced.
        """
        self.stuffWMBS()

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)

        result = jobFactory(min_merge_size=200000, max_merge_size=2000000000,
                            max_merge_events=200000000)

        assert len(result) == 0, \
            "ERROR: No job groups should be returned."

        return

    def testMinMergeSize1a(self):
        """
        _testMinMergeSize1a_

        Set the minimum merge size to be 20,000 bytes which is more than the
        sum of all file sizes in the WMBS instance and mark the fileset as
        closed.  Verify that one job containing all files is pushed out.
        """
        self.stuffWMBS()
        self.mergeFileset.markOpen(False)

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)

        result = jobFactory(min_merge_size=200000, max_merge_size=2000000,
                            max_merge_events=2000000)

        assert len(result) == 1, \
            "ERROR: More than one JobGroup returned: %s" % len(result)

        assert len(result[0].jobs) == 1, \
            "Error: One job should have been returned: %s" % len(result[0].jobs)

        self.assertEqual(result[0].jobs[0]["estimatedDiskUsage"], 10 + 2 * 100)

        self.assertEqual(result[0].jobs[0]["possiblePSN"], set(["T1_US_FNAL"]))

        goldenFiles = ["file1", "file2", "file3", "file4", "fileA", "fileB",
                       "fileC", "fileI", "fileII", "fileIII", "fileIV"]

        jobFiles = result[0].jobs[0].getFiles()

        currentRun = 0
        currentLumi = 0
        currentEvent = 0
        for jobFile in jobFiles:
            jobFile.loadData()
            self.assertTrue(jobFile["lfn"] in goldenFiles,
                            "Error: Unknown file: %s" % jobFile["lfn"])
            goldenFiles.remove(jobFile["lfn"])

            fileRun = list(jobFile["runs"])[0].run
            fileLumi = min(list(jobFile["runs"])[0])
            fileEvent = jobFile["first_event"]

            if currentRun == 0:
                currentRun = fileRun
                currentLumi = fileLumi
                currentEvent = fileEvent
                continue

            assert fileRun >= currentRun, \
                "ERROR: Files not sorted by run."

            if fileRun == currentRun:
                assert fileLumi >= currentLumi, \
                    "ERROR: Files not ordered by lumi"

            if fileLumi == currentLumi:
                assert fileEvent >= currentEvent, \
                    "ERROR: Files not ordered by first event"

            currentRun = fileRun
            currentLumi = fileLumi
            currentEvent = fileEvent

        return

    def testMaxMergeSize(self):
        """
        _testMaxMergeSize_

        Set the maximum merge size to be 100000 bytes.  Verify that two merge
        jobs are created, one for the one large file and another for the rest of
        the files.  Verify that each merge job contains the expected files and
        that we merge across runs.
        """
        self.stuffWMBS()

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)

        result = jobFactory(min_merge_size=1, max_merge_size=100000,
                            max_merge_events=200000)

        assert len(result) == 1, \
            "ERROR: More than one JobGroup returned: %s" % result

        assert len(result[0].jobs) == 2, \
            "ERROR: Two jobs should have been returned."

        goldenFilesA = ["file1", "file2", "file3", "file4", "fileA", "fileB",
                        "fileC", "fileI", "fileII", "fileIII"]
        goldenFilesB = ["fileIV"]

        for job in result[0].jobs:

            self.assertEqual(job["possiblePSN"], set(["T1_US_FNAL"]))

            jobFiles = job.getFiles()

            if jobFiles[0]["lfn"] in goldenFilesA:
                self.assertEqual(job["estimatedDiskUsage"], 11)
                goldenFiles = goldenFilesA
            elif jobFiles[0]["lfn"] in goldenFilesB:
                self.assertEqual(job["estimatedDiskUsage"], 2 * 100)
                goldenFiles = goldenFilesB

            currentRun = 0
            currentLumi = 0
            currentEvent = 0
            for jobFile in jobFiles:
                self.assertTrue(jobFile["lfn"] in goldenFiles,
                                "Error: Unknown file: %s" % jobFile["lfn"])

                goldenFiles.remove(jobFile["lfn"])

            fileRun = list(jobFile["runs"])[0].run
            fileLumi = min(list(jobFile["runs"])[0])
            fileEvent = jobFile["first_event"]

            if currentRun == 0:
                currentRun = fileRun
                currentLumi = fileLumi
                currentEvent = fileEvent
                continue

            assert fileRun >= currentRun, \
                "ERROR: Files not sorted by run."

            if fileRun == currentRun:
                assert fileLumi >= currentLumi, \
                    "ERROR: Files not ordered by lumi"

                if fileLumi == currentLumi:
                    assert fileEvent >= currentEvent, \
                        "ERROR: Files not ordered by first event"

            currentRun = fileRun
            currentLumi = fileLumi
            currentEvent = fileEvent

        assert len(goldenFilesA) == 0 and len(goldenFilesB) == 0, \
            "ERROR: Files missing from merge jobs."

        return

    def testMaxEvents(self):
        """
        _testMaxEvents_

        Verify the the max_merge_events parameter works and that we correctly
        merge across runs.
        """
        self.stuffWMBS()

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)

        result = jobFactory(min_merge_size=1, max_merge_size=20000000,
                            max_merge_events=100000)

        assert len(result) == 1, \
            "ERROR: More than one JobGroup returned: %s" % result

        assert len(result[0].jobs) == 2, \
            "ERROR: Two jobs should have been returned: %s" % len(result[0].jobs)

        goldenFilesA = ["file1", "file2", "file3", "file4", "fileA", "fileB",
                        "fileC", "fileI", "fileII", "fileIV"]
        goldenFilesB = ["fileIII"]

        for job in result[0].jobs:

            self.assertEqual(job["possiblePSN"], set(["T1_US_FNAL"]))

            jobFiles = job.getFiles()

            if jobFiles[0]["lfn"] in goldenFilesA:
                self.assertEqual(job["estimatedDiskUsage"], 9 + 2 * 100)
                goldenFiles = goldenFilesA
            elif jobFiles[0]["lfn"] in goldenFilesB:
                self.assertEqual(job["estimatedDiskUsage"], 2)
                goldenFiles = goldenFilesB

            currentRun = 0
            currentLumi = 0
            currentEvent = 0
            for jobFile in jobFiles:
                self.assertTrue(jobFile["lfn"] in goldenFiles,
                                "Error: Unknown file: %s" % jobFile["lfn"])

                goldenFiles.remove(jobFile["lfn"])

                fileRun = list(jobFile["runs"])[0].run
                fileLumi = min(list(jobFile["runs"])[0])
                fileEvent = jobFile["first_event"]

                if currentRun == 0:
                    currentRun = fileRun
                    currentLumi = fileLumi
                    currentEvent = fileEvent
                    continue

                assert fileRun >= currentRun, \
                    "ERROR: Files not sorted by run: %s, %s" % (fileRun, currentRun)

                if fileRun == currentRun:
                    assert fileLumi >= currentLumi, \
                        "ERROR: Files not ordered by lumi"

                    if fileLumi == currentLumi:
                        assert fileEvent >= currentEvent, \
                            "ERROR: Files not ordered by first event"

                currentRun = fileRun
                currentLumi = fileLumi
                currentEvent = fileEvent

        assert len(goldenFilesA) == 0 and len(goldenFilesB) == 0 and \
               "ERROR: Files missing from merge jobs."

        return

    def testMinMergeSize1aNoRunMerge(self):
        """
        _testMinMergeSize1aNoRunMerge_

        Set the minimum merge size to be 20,000 bytes which is more than the
        sum of all file sizes in the WMBS instance and mark the fileset as
        closed.  Verify that two jobs are pushed out and that we don't merge
        accross run boundaries.
        """
        self.stuffWMBS()
        self.mergeFileset.markOpen(False)

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)

        result = jobFactory(min_merge_size=200000, max_merge_size=2000000,
                            max_merge_events=2000000, merge_across_runs=False)

        assert len(result) == 1, \
            "ERROR: More than one JobGroup returned: %s" % len(result)

        assert len(result[0].jobs) == 2, \
            "Error: Two jobs should have been returned: %s" % len(result[0].jobs)

        goldenFilesA = ["file1", "file2", "file3", "file4",
                        "fileA", "fileB", "fileC"]
        goldenFilesB = ["fileI", "fileII", "fileIII", "fileIV"]
        goldenFilesA.sort()
        goldenFilesB.sort()

        for job in result[0].jobs:

            self.assertEqual(job["possiblePSN"], set(["T1_US_FNAL"]))

            currentRun = 0
            currentLumi = 0
            currentEvent = 0
            jobLFNs = []

            for jobFile in job.getFiles():
                jobFile.loadData()
                jobLFNs.append(jobFile["lfn"])

                fileRun = list(jobFile["runs"])[0].run
                fileLumi = min(list(jobFile["runs"])[0])
                fileEvent = jobFile["first_event"]

                if currentRun == 0:
                    currentRun = fileRun
                    currentLumi = fileLumi
                    currentEvent = fileEvent
                    continue

                assert fileRun >= currentRun, \
                    "ERROR: Files not sorted by run."

                if fileRun == currentRun:
                    assert fileLumi >= currentLumi, \
                        "ERROR: Files not ordered by lumi"

                if fileLumi == currentLumi:
                    assert fileEvent >= currentEvent, \
                        "ERROR: Files not ordered by first event"

                currentRun = fileRun
                currentLumi = fileLumi
                currentEvent = fileEvent

            jobLFNs.sort()
            if jobLFNs == goldenFilesA:
                self.assertEqual(job["estimatedDiskUsage"], 8)
                goldenFilesA = []
            else:
                self.assertEqual(job["estimatedDiskUsage"], 3 + 2 * 100)
                self.assertEqual(jobLFNs, goldenFilesB,
                                 "Error: LFNs do not match.")
                goldenFilesB = []

        return

    def testMaxMergeSizeNoRunMerge(self):
        """
        _testMaxMergeSizeNoRunMerge_

        Set the maximum merge size to be 100000 bytes.  Verify that two merge
        jobs are created, one for the one large file and another for the rest of
        the files.  Verify that each merge job contains the expected files and
        that we don't merge across run boundaries.
        """
        self.stuffWMBS()

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)

        result = jobFactory(min_merge_size=1, max_merge_size=100000,
                            max_merge_events=200000, merge_across_runs=False)

        assert len(result) == 1, \
            "ERROR: More than one JobGroup returned: %s" % result

        assert len(result[0].jobs) == 3, \
            "ERROR: Three jobs should have been returned."

        goldenFilesA = ["file1", "file2", "file3", "file4", "fileA", "fileB",
                        "fileC"]
        goldenFilesB = ["fileI", "fileII", "fileIII"]
        goldenFilesC = ["fileIV"]

        for job in result[0].jobs:

            self.assertEqual(job["possiblePSN"], set(["T1_US_FNAL"]))

            jobFiles = job.getFiles()

            if jobFiles[0]["lfn"] in goldenFilesA:
                self.assertEqual(job["estimatedDiskUsage"], 8)
                goldenFiles = goldenFilesA
            elif jobFiles[0]["lfn"] in goldenFilesB:
                self.assertEqual(job["estimatedDiskUsage"], 4)
                goldenFiles = goldenFilesB
            else:
                self.assertEqual(job["estimatedDiskUsage"], 2 * 100)
                goldenFiles = goldenFilesC

            currentRun = 0
            currentLumi = 0
            currentEvent = 0
            for jobFile in jobFiles:
                self.assertTrue(jobFile["lfn"] in goldenFiles,
                                "Error: Unknown file: %s" % jobFile["lfn"])

                goldenFiles.remove(jobFile["lfn"])

            fileRun = list(jobFile["runs"])[0].run
            fileLumi = min(list(jobFile["runs"])[0])
            fileEvent = jobFile["first_event"]

            if currentRun == 0:
                currentRun = fileRun
                currentLumi = fileLumi
                currentEvent = fileEvent
                continue

            self.assertTrue(fileRun >= currentRun,
                            "ERROR: Files not sorted by run.")
            if fileRun == currentRun:
                self.assertTrue(fileLumi >= currentLumi,
                                "ERROR: Files not ordered by lumi")
                if fileLumi == currentLumi:
                    self.assertTrue(fileEvent >= currentEvent,
                                    "ERROR: Files not ordered by first event")

            currentRun = fileRun
            currentLumi = fileLumi
            currentEvent = fileEvent

        self.assertTrue(len(goldenFilesA) == 0 and len(goldenFilesB) == 0,
                        "ERROR: Files missing from merge jobs.")

        return

    def testMaxEventsNoRunMerge(self):
        """
        _testMaxEventsNoRunMerge_

        Verify that the max events merge parameter works correctly and that we
        don't merge accross run boundaries.
        """
        self.stuffWMBS()

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)

        result = jobFactory(min_merge_size=1, max_merge_size=20000000,
                            max_merge_events=100000, merge_across_runs=False)

        self.assertTrue(len(result) == 1,
                        "ERROR: More than one JobGroup returned: %s" % result)

        self.assertTrue(len(result[0].jobs) == 3,
                        "ERROR: Three jobs should have been returned: %s" % len(result[0].jobs))

        goldenFilesA = ["file1", "file2", "file3", "file4", "fileA", "fileB",
                        "fileC", ]
        goldenFilesB = ["fileI", "fileII", "fileIV"]
        goldenFilesC = ["fileIII"]

        for job in result[0].jobs:

            self.assertEqual(job["possiblePSN"], set(["T1_US_FNAL"]))

            jobFiles = job.getFiles()

            if jobFiles[0]["lfn"] in goldenFilesA:
                self.assertEqual(job["estimatedDiskUsage"], 8)
                goldenFiles = goldenFilesA
            elif jobFiles[0]["lfn"] in goldenFilesB:
                self.assertEqual(job["estimatedDiskUsage"], 2 + 2 * 100)
                goldenFiles = goldenFilesB
            else:
                self.assertEqual(job["estimatedDiskUsage"], 2 * 1)
                goldenFiles = goldenFilesC

            currentRun = 0
            currentLumi = 0
            currentEvent = 0
            for jobFile in jobFiles:
                self.assertTrue(jobFile["lfn"] in goldenFiles,
                                "Error: Unknown file: %s" % jobFile["lfn"])

                goldenFiles.remove(jobFile["lfn"])

                fileRun = list(jobFile["runs"])[0].run
                fileLumi = min(list(jobFile["runs"])[0])
                fileEvent = jobFile["first_event"]

                if currentRun == 0:
                    currentRun = fileRun
                    currentLumi = fileLumi
                    currentEvent = fileEvent
                    continue

                self.assertTrue(fileRun >= currentRun,
                                "ERROR: Files not sorted by run: %s, %s" % (fileRun, currentRun))
                if fileRun == currentRun:
                    self.assertTrue(fileLumi >= currentLumi,
                                    "ERROR: Files not ordered by lumi")
                    if fileLumi == currentLumi:
                        self.assertTrue(fileEvent >= currentEvent,
                                        "ERROR: Files not ordered by first event")

                currentRun = fileRun
                currentLumi = fileLumi
                currentEvent = fileEvent

        self.assertTrue(len(goldenFilesA) == 0 and len(goldenFilesB) == 0 and len(goldenFilesC) == 0,
                        "ERROR: Files missing from merge jobs.")

        return

    def testLocationMerging(self):
        """
        _testLocationMerging_

        Verify that files residing on different SEs are not merged together in
        the same job.
        """
        self.stuffWMBS()

        locationAction = self.daoFactory(classname = "Locations.New")
        locationAction.execute(siteName = "T1_UK_RAL", pnn = "T1_UK_RAL_Disk")

        fileSite2 = File(lfn = "fileRAL", size = 4098, events = 1024,
                         first_event = 0, locations = set(["T1_UK_RAL_Disk"]))
        fileSite2.addRun(Run(1, *[46]))
        fileSite2.create()

        self.mergeFileset.addFile(fileSite2)
        self.mergeFileset.commit()

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)

        result = jobFactory(min_merge_size=4097, max_merge_size=99999999,
                            max_merge_events=999999999, merge_across_runs=False)

        assert len(result) == 1, \
            "ERROR: More than one JobGroup returned."

        assert len(result[0].jobs) == 3, \
            "ERROR: Three jobs should have been returned."

        ralJobs = 0
        fnalJobs = 0
        for job in result[0].jobs:
            if job["possiblePSN"] == set(["T1_UK_RAL"]):
                ralJobs += 1
            elif job["possiblePSN"] == set(["T1_US_FNAL"]):
                fnalJobs += 1

        self.assertEqual(ralJobs, 1)
        self.assertEqual(fnalJobs, 2)

        return

    def testMaxWaitTime(self):
        """
        _testMaxWaitTime_

        Set the max wait times to be negative - this should force all files to merge
        out immediately

        Using the first setup as the first merge test which should normally produce
        no jobGroups
        """
        self.stuffWMBS()

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)

        result = jobFactory(min_merge_size=200000, max_merge_size=2000000000,
                            max_merge_events=200000000, max_wait_time=-10)

        # Everything should be in one, small jobGroup
        self.assertEqual(len(result), 1)
        self.assertEqual(len(result[0].jobs), 1)
        job = result[0].jobs[0]
        # All files should be in one job
        self.assertEqual(len(job.getFiles()), 11)

        return

    def testDifferentSubscritionIDs(self):
        """
        _testDifferentSubscriptionIDs_

        Make sure that the merge splitting still runs if the subscription ID
        is not equal to the workflow ID.
        """
        myThread = threading.currentThread()
        myThread.transaction.begin()
        dummyWorkflow = Workflow(name="dummyWorkflow", spec="bunk49",
                                 owner="Steve", task="Test2")
        dummyWorkflow.create()
        dummyFileset = Fileset(name="dummyFileset")
        dummyFileset.create()
        dummySubscription1 = Subscription(fileset=dummyFileset,
                                          workflow=dummyWorkflow,
                                          split_algo="ParentlessMergeBySize")
        dummySubscription2 = Subscription(fileset=dummyFileset,
                                          workflow=dummyWorkflow,
                                          split_algo="ParentlessMergeBySize")
        dummySubscription1.create()
        dummySubscription2.create()
        myThread.transaction.commit()

        self.stuffWMBS()
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.mergeSubscription)
        result = jobFactory(min_merge_size=4097, max_merge_size=99999999,
                            max_merge_events=999999999, merge_across_runs=False)
        self.assertEqual(len(result), 1)
        jobGroup = result[0]
        self.assertEqual(len(jobGroup.jobs), 2)
        return


if __name__ == '__main__':
    unittest.main()



"""
__ListForJobUpdater__

MySQL implementation of ListForJobUpdater

Created on Apr 16, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class ListForJobUpdater(DBFormatter):
    """
    _ListForJobUpdater_

    List the available workflows and tasks in WMBS with the workflow
    priority and the sub task priority
    """
    sql = """SELECT wmbs_workflow.name, wmbs_workflow.task,
                    wmbs_workflow.priority AS workflow_priority,
                    MIN(wmbs_sub_types.priority) AS task_priority
             FROM wmbs_workflow
             INNER JOIN wmbs_subscription ON
               wmbs_workflow.id = wmbs_subscription.workflow
             INNER JOIN wmbs_sub_types ON
               wmbs_subscription.subtype = wmbs_sub_types.id
             GROUP BY wmbs_workflow.name, wmbs_workflow.task,
                      wmbs_workflow.priority"""

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql,
                                      conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
Test case for CRIC
"""
from __future__ import print_function, division

import unittest

from nose.plugins.attrib import attr

from Utils.PythonVersion import PY3

from WMCore.Services.CRIC.CRIC import CRIC
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase


class CRICTest(EmulatedUnitTestCase):
    """
    Unit tests for CRIC Services module
    """

    def __init__(self, methodName='runTest'):
        super(CRICTest, self).__init__(methodName=methodName)

    def setUp(self):
        """
        Setup for unit tests
        """
        super(CRICTest, self).setUp()
        self.myCRIC = CRIC()
        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def testConfig(self):
        """
        Test service attributes and the override mechanism
        """
        self.assertEqual(self.myCRIC['endpoint'], 'https://cms-cric.cern.ch/')
        self.assertEqual(self.myCRIC['cacheduration'], 1)
        self.assertEqual(self.myCRIC['accept_type'], 'application/json')
        self.assertEqual(self.myCRIC['content_type'], 'application/json')
        self.assertEqual(self.myCRIC['requests'].pycurl, True)

        newParams = {"cacheduration": 100, "pycurl": False}
        cric = CRIC(url='https://BLAH.cern.ch/', configDict=newParams)
        self.assertEqual(cric['endpoint'], 'https://BLAH.cern.ch/')
        self.assertEqual(cric['cacheduration'], newParams['cacheduration'])
        self.assertEqual(cric['requests'].pycurl, False)

    @attr('integration')
    def testWhoAmI(self):
        """
        Test user mapping information from the request headers
        """
        print("This test only works with service certificates, not user proxies")
        self.assertTrue(self.myCRIC.whoAmI())  # empty list if nothing

    def testUserNameDN(self):
        """
        Tests user name to DN lookup
        """
        expectedDN = "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=amaltaro/CN=718748/CN=Alan Malta Rodrigues"
        username = "amaltaro"
        dn = self.myCRIC.userNameDn(username)
        self.assertEqual(dn, expectedDN)

    def testUserNameDNWithApostrophe(self):
        """
        Test DN lookup with an apostrophe
        """
        expectedDN = "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=gdimperi/CN=728001/CN=Giulia D'Imperio"
        username = "gdimperi"
        dn = self.myCRIC.userNameDn(username)
        self.assertEqual(dn, expectedDN)

    def testGetAllPSNs(self):
        """
        Test API which fetches all PSNs
        """
        result = self.myCRIC.getAllPSNs()
        self.assertNotIn('T1_US_FNAL_Disk', result)
        self.assertNotIn('T1_US_FNAL_MSS', result)
        self.assertIn('T1_US_FNAL', result)
        self.assertIn('T2_CH_CERN', result)
        self.assertIn('T2_CH_CERN_HLT', result)

        t1s = [psn for psn in result if psn.startswith('T1_')]
        self.assertTrue(len(t1s) < 10)

        t2s = [psn for psn in result if psn.startswith('T2_')]
        self.assertTrue(len(t2s) > 30)

        t3s = [psn for psn in result if psn.startswith('T3_')]
        self.assertTrue(len(t3s) > 10)

        self.assertTrue(len(result) > 70)

        return

    def testGetAllPhEDExNodeNames(self):
        """
        Test API to get all PhEDEx Node Names
        """
        result = self.myCRIC.getAllPhEDExNodeNames(excludeBuffer=True)
        self.assertFalse([pnn for pnn in result if pnn.endswith('_Buffer')])

        result = self.myCRIC.getAllPhEDExNodeNames(excludeBuffer=False)
        self.assertTrue(len([pnn for pnn in result if pnn.endswith('_Buffer')]) > 5)

        result = self.myCRIC.getAllPhEDExNodeNames(pattern='T1.*', excludeBuffer=True)
        self.assertFalse([pnn for pnn in result if not pnn.startswith('T1_')])
        self.assertTrue(len(result) > 10)

        result = self.myCRIC.getAllPhEDExNodeNames(excludeBuffer=True)
        self.assertTrue([pnn for pnn in result if pnn.startswith('T1_')])
        self.assertTrue([pnn for pnn in result if pnn.startswith('T2_')])
        self.assertTrue([pnn for pnn in result if pnn.startswith('T3_')])
        self.assertTrue(len(result) > 60)

        result1 = self.myCRIC.getAllPhEDExNodeNames(pattern='.*', excludeBuffer=True)
        self.assertTrue([pnn for pnn in result1 if pnn.startswith('T1_')])
        self.assertTrue([pnn for pnn in result1 if pnn.startswith('T2_')])
        self.assertTrue([pnn for pnn in result1 if pnn.startswith('T3_')])
        self.assertTrue(len(result) > 60)

        self.assertItemsEqual(result, result1)

        # test a few PSNs
        self.assertNotIn('T1_US_FNAL', result)
        self.assertNotIn('T2_CH_CERN_HLT', result)

        # test a few PNNs
        self.assertIn('T1_US_FNAL_Disk', result)
        self.assertIn('T1_US_FNAL_MSS', result)
        self.assertIn('T2_CH_CERN', result)

        return

    def testPNNstoPSNs(self):
        """
        Test mapping PhEDEx Node Names to Processing Site Names
        """
        self.assertEqual(self.myCRIC.PNNstoPSNs([]), [])

        self.assertItemsEqual(self.myCRIC.PNNstoPSNs(['T1_US_FNAL_MSS']), [])
        self.assertItemsEqual(self.myCRIC.PNNstoPSNs(['T1_US_FNAL_Disk']), ['T1_US_FNAL'])

        pnns = ['T1_US_FNAL_Disk', 'T1_US_FNAL_Buffer', 'T1_US_FNAL_MSS']
        self.assertItemsEqual(self.myCRIC.PNNstoPSNs(pnns), ['T1_US_FNAL'])

        pnns = ['T2_CH_CERN', 'T2_DE_DESY']
        psns = ['T2_CH_CERN_HLT', 'T2_DE_DESY', 'T2_CH_CERN']
        self.assertItemsEqual(self.myCRIC.PNNstoPSNs(pnns), psns)

        psns = ['T2_UK_London_IC', 'T3_UK_London_QMUL', 'T3_UK_SGrid_Oxford', 'T3_UK_London_RHUL', 'T3_UK_ScotGrid_GLA']
        self.assertItemsEqual(self.myCRIC.PNNstoPSNs('T2_UK_London_IC'), psns)

        pnns = ['T2_UK_London_IC', 'T2_US_Purdue']
        self.assertItemsEqual(self.myCRIC.PNNstoPSNs(pnns), psns + ['T2_US_Purdue'])

        return

    def testPSNstoPNNs(self):
        """
        Test mapping Processing Site Names to PhEDEx Node Names
        """
        self.assertEqual(self.myCRIC.PSNstoPNNs([]), [])

        # test a few PNNs
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs('T1_US_FNAL_Disk'), [])
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs(['T1_DE_KIT_MSS']), [])

        # test a few PSNs
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs(['T1_US_FNAL']), ['T1_US_FNAL_Disk', 'T3_US_FNALLPC'])
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs(['T2_CH_CERN_HLT']), ['T2_CH_CERN'])

        pnns = ['T2_CH_CERN', 'T2_CH_CERNBOX']
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs(['T2_CH_CERN']), pnns)
        psns = ['T2_CH_CERN', 'T2_CH_CERN_HLT']
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs(psns), pnns)

        self.assertItemsEqual(self.myCRIC.PSNstoPNNs(['T2_UK_London_IC']), ['T2_UK_London_IC'])
        pnns = ['T2_UK_London_IC', 'T2_UK_London_Brunel', 'T2_UK_SGrid_RALPP']
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs('T3_UK_London_QMUL'), pnns)
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs('T3_UK_ScotGrid_GLA'), pnns)
        pnns = ['T2_UK_London_IC', 'T2_UK_SGrid_RALPP']
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs('T3_UK_SGrid_Oxford'), pnns)
        self.assertItemsEqual(self.myCRIC.PSNstoPNNs('T3_UK_London_RHUL'), pnns)

        self.assertItemsEqual(self.myCRIC.PSNstoPNNs(['T2_US_Purdue']), ['T2_US_Purdue'])

        return

    def testPSNtoPNNMap(self):
        """
        Test API to get a map of PSNs to PNNs
        """
        with self.assertRaises(TypeError):
            self.myCRIC.PSNtoPNNMap(1)
        with self.assertRaises(TypeError):
            self.myCRIC.PSNtoPNNMap({'config': 'blah'})
        with self.assertRaises(TypeError):
            self.myCRIC.PSNtoPNNMap(['blah'])

        result = self.myCRIC.PSNtoPNNMap()
        self.assertTrue([psn for psn in result.keys() if psn.startswith('T1_')])
        self.assertTrue([psn for psn in result.keys() if psn.startswith('T2_')])
        self.assertTrue([psn for psn in result.keys() if psn.startswith('T3_')])
        self.assertTrue(len(result) > 50)

        result = self.myCRIC.PSNtoPNNMap(psnPattern='T1.*')
        self.assertFalse([psn for psn in result.keys() if not psn.startswith('T1_')])
        self.assertTrue(len(result) < 10)

        result = self.myCRIC.PSNtoPNNMap(psnPattern='T2.*')
        self.assertFalse([psn for psn in result.keys() if not psn.startswith('T2_')])
        self.assertTrue(len(result) > 10)

        result = self.myCRIC.PSNtoPNNMap(psnPattern='T3.*')
        self.assertFalse([psn for psn in result.keys() if not psn.startswith('T3_')])
        self.assertTrue(len(result) > 10)

        result = self.myCRIC.PSNtoPNNMap('T2_CH_CERN$')
        self.assertItemsEqual(list(result), ['T2_CH_CERN'])
        self.assertItemsEqual(result['T2_CH_CERN'], ['T2_CH_CERNBOX', 'T2_CH_CERN'])

        # test a exact site name, which is treated as a regex and yields a confusing result!!!
        result = self.myCRIC.PSNtoPNNMap('T2_CH_CERN')
        self.assertItemsEqual(list(result), ['T2_CH_CERN', 'T2_CH_CERN_HLT'])
        self.assertItemsEqual(result['T2_CH_CERN'], ['T2_CH_CERNBOX', 'T2_CH_CERN'])
        self.assertItemsEqual(result['T2_CH_CERN_HLT'], ['T2_CH_CERN'])

        result = self.myCRIC.PSNtoPNNMap('T2_CH_CERN_HLT')
        self.assertItemsEqual(list(result), ['T2_CH_CERN_HLT'])
        self.assertItemsEqual(result['T2_CH_CERN_HLT'], ['T2_CH_CERN'])

        result = self.myCRIC.PSNtoPNNMap('T1_US_FNAL')
        self.assertItemsEqual(list(result), ['T1_US_FNAL'])
        self.assertItemsEqual(result['T1_US_FNAL'], ['T1_US_FNAL_Disk', 'T3_US_FNALLPC'])

        # test a PNN as input, expecting nothing mapped to it
        self.assertItemsEqual(self.myCRIC.PSNtoPNNMap('T1_US_FNAL_Disk'), {})

        return

    def testPNNtoPSNMap(self):
        """
        Test API to get a map of PSNs to PNNs
        """
        with self.assertRaises(TypeError):
            self.myCRIC.PNNtoPSNMap(1)
        with self.assertRaises(TypeError):
            self.myCRIC.PNNtoPSNMap({'config': 'blah'})
        with self.assertRaises(TypeError):
            self.myCRIC.PNNtoPSNMap(['blah'])

        result = self.myCRIC.PNNtoPSNMap()
        self.assertTrue([psn for psn in result.keys() if psn.startswith('T1_')])
        self.assertTrue([psn for psn in result.keys() if psn.startswith('T2_')])
        self.assertTrue([psn for psn in result.keys() if psn.startswith('T3_')])
        self.assertTrue(len(result) > 50)

        result = self.myCRIC.PNNtoPSNMap(pnnPattern='T1.*')
        self.assertFalse([psn for psn in result.keys() if not psn.startswith('T1_')])
        self.assertTrue(len(result) < 10)

        result = self.myCRIC.PNNtoPSNMap(pnnPattern='T2.*')
        self.assertFalse([psn for psn in result.keys() if not psn.startswith('T2_')])
        self.assertTrue(len(result) > 10)

        result = self.myCRIC.PNNtoPSNMap(pnnPattern='T3.*')
        self.assertFalse([psn for psn in result.keys() if not psn.startswith('T3_')])
        self.assertTrue(len(result) > 10)

        result = self.myCRIC.PNNtoPSNMap('T2_CH_CERN$')
        self.assertItemsEqual(list(result), ['T2_CH_CERN'])
        self.assertItemsEqual(result['T2_CH_CERN'], ['T2_CH_CERN_HLT', 'T2_CH_CERN'])

        # test a exact site name, which is treated as a regex and yields a confusing result!!!
        result = self.myCRIC.PNNtoPSNMap('T2_CH_CERN')
        self.assertItemsEqual(list(result), ['T2_CH_CERN', 'T2_CH_CERNBOX'])
        self.assertItemsEqual(result['T2_CH_CERN'], ['T2_CH_CERN_HLT', 'T2_CH_CERN'])
        self.assertItemsEqual(result['T2_CH_CERNBOX'], ['T2_CH_CERN'])

        result = self.myCRIC.PNNtoPSNMap('T2_CH_CERN_HLT')
        self.assertItemsEqual(result, {})

        result = self.myCRIC.PNNtoPSNMap('T1_US_FNAL')
        self.assertItemsEqual(list(result), ['T1_US_FNAL_Disk'])
        self.assertItemsEqual(result['T1_US_FNAL_Disk'], ['T1_US_FNAL'])

        return



if __name__ == '__main__':
    unittest.main()



from __future__ import (division, print_function)

import time
from WMCore.REST.CherryPyPeriodicTask import CherryPyPeriodicTask
from WMCore.WMStats.DataStructs.DataCache import DataCache
from WMCore.Services.WMStats.WMStatsReader import WMStatsReader
from WMCore.ReqMgr.DataStructs.RequestStatus import WMSTATS_JOB_INFO, WMSTATS_NO_JOB_INFO

class DataCacheUpdate(CherryPyPeriodicTask):

    def __init__(self, rest, config):
        self.getJobInfo = getattr(config, "getJobInfo", False)

        super(DataCacheUpdate, self).__init__(config)

    def setConcurrentTasks(self, config):
        """
        sets the list of functions which
        """
        self.concurrentTasks = [{'func': self.gatherActiveDataStats, 'duration': 300}]

    def gatherActiveDataStats(self, config):
        """
        gather active data statistics
        """
        self.logger.info("Starting gatherActiveDataStats with jobInfo set to: %s", self.getJobInfo)
        try:
            tStart = time.time()
            if DataCache.islatestJobDataExpired():
                wmstatsDB = WMStatsReader(config.wmstats_url, reqdbURL=config.reqmgrdb_url,
                                          reqdbCouchApp="ReqMgr", logger=self.logger)
                self.logger.info("Getting active data with job info for statuses: %s", WMSTATS_JOB_INFO)
                jobData = wmstatsDB.getActiveData(WMSTATS_JOB_INFO, jobInfoFlag=self.getJobInfo)
                self.logger.info("Getting active data with NO job info for statuses: %s", WMSTATS_NO_JOB_INFO)
                tempData = wmstatsDB.getActiveData(WMSTATS_NO_JOB_INFO, jobInfoFlag=False)
                jobData.update(tempData)
                self.logger.info("Running setlatestJobData...")
                DataCache.setlatestJobData(jobData)
                self.logger.info("DataCache is up-to-date with %d requests data", len(jobData))
        except Exception as ex:
            self.logger.exception("Exception updating DataCache. Error: %s", str(ex))
        self.logger.info("Total time loading data from ReqMgr2 and WMStats: %s", time.time() - tStart)
        return



from WMQuality.Emulators.WMSpecGenerator.WMSpecGenerator import WMSpecGenerator
from WMCore.ReqMgr.DataStructs.RequestStatus import check_allowed_transition

class ReqMgr(dict):

    def __init__(self, *args, **kwargs):
        """
        all the private valuable is defined for test values
        """
        self.specGenerator = WMSpecGenerator()
        self.count = 0
        self.maxWmSpec = kwargs.setdefault('numOfSpecs', 1)
        self.type = kwargs.setdefault("type", 'ReReco')
        if self.type != 'ReReco':
            raise TypeError('unknown request type %s' % self.type)
        self.splitter = kwargs.setdefault('splitter', 'DatasetBlock')
        self.inputDataset = kwargs.setdefault('inputDataset', None)
        self.dbsUrl = kwargs.setdefault('dbsUrl', None)
        self.status = {}
        self.progress = {}
        self.msg = {}
        self.names = []
        self.openRunningTimeout = kwargs.setdefault('openRunningTimeout', 0)
        import logging
        self['logger'] = logging

    def getAssignment(self, teamName=None, request=None):
        if self.count < self.maxWmSpec:
            if self.type == 'ReReco':
                specName = "ReRecoTest_v%sEmulator" % self.count
                specUrl =self.specGenerator.createReRecoSpec(specName, "file",
                                                             self.splitter,
                                                             InputDataset=self.inputDataset,
                                                             DbsUrl=self.dbsUrl,
                                                             OpenRunningTimeout = self.openRunningTimeout)
            self.names.append(specName)
            self.status[specName] = 'assigned'
            #specName = "FakeProductionSpec_%s" % self.count
            #specUrl =self.specGenerator.createProductionSpec(specName, "file")
            #specName = "FakeProcessingSpec_%s" % self.count
            #specUrl =self.specGenerator.createProcessingSpec(specName, "file")

            self.count += 1
            # returns list of list(rquest name, spec url)
            return [[specName, specUrl],]
        else:
            return []

    def getRunningOpen(self, teamName):
        """Returns a list of request names of requests in running-open state"""
        result = []
        for request in self.status:
            if self.status[request] == 'running-open':
                result.append(request)
        return result

    def getRequest(self, requestName):
        """Get request info"""
        if requestName not in self.names:
            raise RuntimeError("unknown request %s" % requestName)

        request = {'RequestName' : requestName,
                   'RequestStatus' : self.status[requestName],
                   'RequestPriority' : 100}
        if requestName in self.progress:
            request.update(self.progress[requestName])
        request.setdefault('percent_complete', 0)
        request.setdefault('percent_success', 0)
        return request

    def putWorkQueue(self, reqName, prodAgentUrl=None):
        self.status[reqName] = 'acquired'

    def reportRequestStatus(self, name, status):
        if not check_allowed_transition(self.status[name], status):
            raise RuntimeError("Invalid status move: %s" % status)
        self.status[name] = status

    def reportRequestProgress(self, name, **args):
        self.progress.setdefault(name, {})
        self.progress[name].update(args)

    def sendMessage(self, request, msg):
        self.msg[request] = msg

    def _removeSpecs(self):
        """
        This is just for clean up not part of emulated function
        """
        self.specGenerator.removeSpecs()

    def getTeam(self):
        """Give a fake team name"""
        return ["The A-Team", "some other bloke"]

    def getRequestByStatus(self, statusList, detail=True):
        # FIXME: this method needs to be fixed accordingly
        if statusList == 'staged':
            specName = "ReRecoTest_v%sEmulator" % self.count
            specUrl =self.specGenerator.createReRecoSpec(specName, "file",
                                                         self.splitter,
                                                         assignKwargs={'SiteWhitelist': ['T2_XX_SiteA']})
            self.names.append(specName)
            self.status[specName] = 'staged'
            #specName = "FakeProductionSpec_%s" % self.count
            #specUrl =self.specGenerator.createProductionSpec(specName, "file")
            #specName = "FakeProcessingSpec_%s" % self.count
            #specUrl =self.specGenerator.createProcessingSpec(specName, "file")

            self.count += 1
            # returns list of list(rquest name, spec url)
            return [[specName, specUrl],]
        else:
            return []


#!/usr/bin/env python
"""
_BasicNaming_

Default name generator using a vaguely sensible convention.
Uses GUIDs to avoid having to keep state

"""
from WMCore.Services.UUIDLib import makeUUID
from WMCore.JobSplitting.Generators.GeneratorInterface import GeneratorInterface



class BasicNaming(GeneratorInterface):
    """
    _BasicNaming_

    Basic task & guid based name generator

    """

    def __call__(self, wmbsJob):
        wmbsJob['id'] = "%s/%s" % (self.task.getPathName(), makeUUID())
        wmbsJob['name'] = "%s/%s" % (self.task.getPathName(), makeUUID())



#!/usr/bin/env python
"""
_New_

MySQL implementation of JobGroup.SetSite
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

from future.utils import listvalues

class GetSite(DBFormatter):
    sql = """SELECT site_name FROM wmbs_location
              WHERE ID = (SELECT location FROM wmbs_jobgroup jg WHERE jg.ID = :jobGroupID)"""


    def format(self, result):
        rowProxy = result[0].fetchall()[0]
        return listvalues(rowProxy)[0]

    def execute(self, jobGroupID = None, conn = None,
                transaction = False):
        binds = {'jobGroupID' : jobGroupID}
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        return self.format(result)



"""
Hello world example using WMCore.REST handling framework.
Info class giving information about ReqMgr database.
Teams, Groups, Software versions handling for ReqMgr.

"""
import cherrypy
from WMCore.REST.Server import RESTEntity, restcall, rows
from WMCore.REST.Tools import tools
from WMCore.Services.WMStats.WMStatsReader import WMStatsReader

from WMCore.REST.Format import PrettyJSONFormat, PrettyJSONHTMLFormat, JSONFormat

class RequestInfo(RESTEntity):
    """
    This class need to move under WMStats server when wmstats server created
    """
    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)
        wmstats_url = "%s/%s" % (self.config.couch_host, self.config.couch_wmstats_db)
        reqdb_url = "%s/%s" % (self.config.couch_host, self.config.couch_reqmgr_db)
        self.wmstats = WMStatsReader(wmstats_url, reqdbURL=reqdb_url, reqdbCouchApp="ReqMgr")

    def validate(self, apiobj, method, api, param, safe):
        args_length = len(param.args)
        if args_length == 1:
            safe.args.append(param.args[0])
            param.args.pop()
        return


    @restcall(formats = [('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self, request_name):
        result = self.wmstats.getRequestSummaryWithJobInfo(request_name)
        return rows([result])


class FinishedStatusInfo(RESTEntity):
    """
    This class need to move under WMStats server when wmstats server created
    """
    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)
        wmstats_url = "%s/%s" % (self.config.couch_host, self.config.couch_wmstats_db)
        reqdb_url = "%s/%s" % (self.config.couch_host, self.config.couch_reqmgr_db)
        self.wmstats = WMStatsReader(wmstats_url, reqdbURL=reqdb_url, reqdbCouchApp="ReqMgr")

    def validate(self, apiobj, method, api, param, safe):
        args_length = len(param.args)
        if args_length == 1:
            safe.args.append(param.args[0])
            param.args.pop()
        return


    @restcall(formats = [('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self, request_name):

        try:
            result = self.wmstats.isWorkflowCompletedWithLogCollectAndCleanUp(request_name)
        except KeyError:
            raise cherrypy.HTTPError(404, "Cannot find request: %s" % request_name)

        return rows([result])

class JobDetailInfo(RESTEntity):
    """
    This class need to move under WMStats server when wmstats server created
    """
    def __init__(self, app, api, config, mount, t0flag=False):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)
        wmstats_url = "%s/%s" % (self.config.couch_host, self.config.couch_wmstats_db)
        reqdb_url = "%s/%s" % (self.config.couch_host, self.config.couch_reqmgr_db)
        if t0flag:
            couchAppName = "T0Request"
        else:
            couchAppName = "ReqMgr"
        self.wmstats = WMStatsReader(wmstats_url, reqdbURL=reqdb_url, reqdbCouchApp=couchAppName)

    def validate(self, apiobj, method, api, param, safe):
        args_length = len(param.args)
        if args_length == 1:
            safe.args.append(param.args[0])
            param.args.pop()

        prop = 'sample_size'
        safe.kwargs[prop] = int(param.kwargs.get(prop, 1))
        if prop in param.kwargs:
            del param.kwargs[prop]
        return


    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('text/html', PrettyJSONHTMLFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self, request_name, sample_size):

        result = self.wmstats.getTaskJobSummaryByRequest(request_name, sample_size)
        return rows([result])

class TeamInfo(RESTEntity):
    """
    This class need to move under WMStats server when wmstats server created
    """
    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)
        wmstats_url = "%s/%s" % (self.config.couch_host, self.config.couch_wmstats_db)
        self.wmstats = WMStatsReader(wmstats_url)

    def validate(self, apiobj, method, api, param, safe):
        args_length = len(param.args)
        if args_length == 1:
            safe.args.append(param.args[0])
            param.args.pop()
        return


    @restcall(formats = [('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self):
        result = self.wmstats.agentsByTeam(filterDrain=False)
        return rows(result)



#!/usr/bin/env python
# encoding: utf-8
"""
AutomaticSeeding_t.py

Created by Dave Evans on 2010-08-30.
Copyright (c) 2010 Fermilab. All rights reserved.
"""
from __future__ import print_function

import sys
import os
import unittest


from WMCore.JobSplitting.Generators.AutomaticSeeding import AutomaticSeeding
from WMCore.DataStructs.Job import Job
from PSetTweaks.PSetTweak import PSetTweak

class AutomaticSeeding_tTests(unittest.TestCase):


    def testA(self):
        """test creating the plugin"""
        try:
            seeder = AutomaticSeeding()
        except Exception as ex:
            msg = "Failed to instantiate an AutomaticSeeder: "
            msg += str(ex)
            self.fail(msg)



    def testB(self):
        """test plugin acts on a Job as expected"""

        job = Job("TestJob")
        seeder = AutomaticSeeding()
        seeder(job)

    def testC(self):
        """test building a tweak from the seeds"""
        job = Job("TestJob")
        seeder = AutomaticSeeding()

        job.addBaggageParameter("process.RandomNumberGeneratorService.seed1.initialSeed", 123445)
        job.addBaggageParameter("process.RandomNumberGeneratorService.seed2.initialSeed", 123445)
        job.addBaggageParameter("process.RandomNumberGeneratorService.seed3.initialSeed", 7464738)
        job.addBaggageParameter("process.RandomNumberGeneratorService.seed44.initialSeed", 98273762)


        seeder(job)

        tweak = PSetTweak()
        for x in job.baggage.process.RandomNumberGeneratorService:
            parameter = "process.RandomNumberGeneratorService.%s.initialSeed" % x._internal_name
            tweak.addParameter(parameter, x.initialSeed)
        print(tweak)




if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# pylint: disable=E1101, C0103, W0401, E1103
# W0401: I am not going to import all those functions by hand
"""

DBSUpload test TestDBSUpload module and the harness

"""
from __future__ import division

import json
import os
import threading
import time
import unittest
from tempfile import mkstemp

from dbs.apis.dbsClient import DbsApi
from nose.plugins.attrib import attr

from WMComponent.DBS3Buffer.DBSBufferBlock import DBSBufferBlock
from WMComponent.DBS3Buffer.DBSBufferDataset import DBSBufferDataset
from WMComponent.DBS3Buffer.DBSBufferFile import DBSBufferFile
from WMComponent.DBS3Buffer.DBSBufferUtil import DBSBufferUtil
from WMComponent.DBS3Buffer.DBSUploadPoller import DBSUploadPoller, parseDBSException
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.Services.UUIDLib import makeUUID
from WMQuality.Emulators import EmulatorSetup
from WMQuality.Emulators.DBSClient.DBS3API import DbsApi as MockDbsApi
from WMQuality.TestInit import TestInit


class MyTestException(Exception):

    def __init__(self, errorMsg=None):
        super(MyTestException, self).__init__(errorMsg)
        if errorMsg is not None:
            self.reason = errorMsg


class DBSUploadTest(unittest.TestCase):
    """
    TestCase for DBSUpload module

    """

    def setUp(self):
        """
        _setUp_

        setUp function for unittest

        """

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMComponent.DBS3Buffer"],
                                useDefault=False)
        self.testDir = self.testInit.generateWorkDir(deleteOnDestruction=False)
        self.configFile = EmulatorSetup.setupWMAgentConfig()

        myThread = threading.currentThread()
        self.bufferFactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                        logger=myThread.logger,
                                        dbinterface=myThread.dbi)

        locationAction = self.bufferFactory(classname="DBSBufferFiles.AddLocation")
        locationAction.execute(siteName="se1.cern.ch")
        locationAction.execute(siteName="se1.fnal.gov")
        locationAction.execute(siteName="malpaquet")
        self.dbsUrl = "https://localhost:1443/dbs/dev/global/DBSWriter"
        self.dbsApi = None
        return

    def tearDown(self):
        """
        _tearDown_

        tearDown function for unittest
        """
        self.testInit.clearDatabase()
        self.testInit.delWorkDir()
        EmulatorSetup.deleteConfig(self.configFile)

        return

    def getConfig(self):
        """
        _getConfig_

        This creates the actual config file used by the component.
        """
        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        # First the general stuff
        config.section_("General")
        config.General.workDir = os.getenv("TESTDIR", os.getcwd())

        config.section_("Agent")
        config.Agent.componentName = 'DBSUpload'
        config.Agent.useHeartbeat = False

        # Now the CoreDatabase information
        # This should be the dialect, dburl, etc
        config.section_("CoreDatabase")
        config.CoreDatabase.connectUrl = os.getenv("DATABASE")
        config.CoreDatabase.socket = os.getenv("DBSOCK")

        config.component_("DBS3Upload")
        config.DBS3Upload.pollInterval = 10
        config.DBS3Upload.logLevel = 'DEBUG'
        # config.DBS3Upload.dbsUrl           = "https://cmsweb-testbed.cern.ch/dbs/dev/global/DBSWriter"
        # config.DBS3Upload.dbsUrl           = "https://dbs3-dev01.cern.ch/dbs/prod/global/DBSWriter"
        config.DBS3Upload.dbsUrl = self.dbsUrl
        config.DBS3Upload.namespace = 'WMComponent.DBS3Buffer.DBS3Upload'
        config.DBS3Upload.componentDir = os.path.join(os.getcwd(), 'Components')
        config.DBS3Upload.nProcesses = 1
        config.DBS3Upload.dbsWaitTime = 0.1
        config.DBS3Upload.datasetType = "VALID"

        # added to skip the StepChain parentage setting test
        config.component_("Tier0Feeder")
        return config

    def createParentFiles(self, acqEra, nFiles=10,
                          workflowName='TestWorkload',
                          taskPath='/TestWorkload/DataTest'):
        """
        _createParentFiles_

        Create several parentless files in DBSBuffer.  This simulates raw files
        in the T0.
        """
        workflowId = self.injectWorkflow(workflowName=workflowName,
                                         taskPath=taskPath)
        parentlessFiles = []

        baseLFN = "/store/data/%s/Cosmics/RAW/v1/000/143/316/" % acqEra
        for i in range(nFiles):
            testFile = DBSBufferFile(lfn=baseLFN + makeUUID() + ".root", size=1024,
                                     events=20, checksums={"cksum": 1},
                                     workflowId=workflowId)
            testFile.setAlgorithm(appName="cmsRun", appVer="CMSSW_3_1_1",
                                  appFam="RAW", psetHash="GIBBERISH",
                                  configContent="MOREGIBBERISH")
            testFile.setDatasetPath("/Cosmics/%s-v1/RAW" % acqEra)

            testFile['block_close_max_wait_time'] = 1000000
            testFile['block_close_max_events'] = 1000000
            testFile['block_close_max_size'] = 1000000
            testFile['block_close_max_files'] = 1000000

            lumis = []
            for j in range(10):
                lumis.append((i * 10) + j)
            testFile.addRun(Run(143316, *lumis))

            testFile.setAcquisitionEra(acqEra)
            testFile.setProcessingVer("1")
            testFile.setGlobalTag("START54::All")
            testFile.create()
            testFile.setLocation("malpaquet")
            parentlessFiles.append(testFile)

        return parentlessFiles

    def createFilesWithChildren(self, moreParentFiles, acqEra):
        """
        _createFilesWithChildren_

        Create several parentless files and then create child files.
        """
        parentFiles = []
        childFiles = []

        baseLFN = "/store/data/%s/Cosmics/RAW/v1/000/143/316/" % acqEra
        for i in range(10):
            testFile = DBSBufferFile(lfn=baseLFN + makeUUID() + ".root", size=1024,
                                     events=20, checksums={"cksum": 1})
            testFile.setAlgorithm(appName="cmsRun", appVer="CMSSW_3_1_1",
                                  appFam="RAW", psetHash="GIBBERISH",
                                  configContent="MOREGIBBERISH")
            testFile.setDatasetPath("/Cosmics/%s-v1/RAW" % acqEra)

            testFile['block_close_max_wait_time'] = 1000000
            testFile['block_close_max_events'] = 1000000
            testFile['block_close_max_size'] = 1000000
            testFile['block_close_max_files'] = 1000000

            lumis = []
            for j in range(10):
                lumis.append((i * 10) + j)
            testFile.addRun(Run(143316, *lumis))

            testFile.setAcquisitionEra(acqEra)
            testFile.setProcessingVer("1")
            testFile.setGlobalTag("START54::All")
            testFile.create()
            testFile.setLocation("malpaquet")
            parentFiles.append(testFile)

        baseLFN = "/store/data/%s/Cosmics/RECO/v1/000/143/316/" % acqEra
        for i in range(5):
            testFile = DBSBufferFile(lfn=baseLFN + makeUUID() + ".root", size=1024,
                                     events=20, checksums={"cksum": 1})
            testFile.setAlgorithm(appName="cmsRun", appVer="CMSSW_3_1_1",
                                  appFam="RECO", psetHash="GIBBERISH",
                                  configContent="MOREGIBBERISH")
            testFile.setDatasetPath("/Cosmics/%s-v1/RECO" % acqEra)

            testFile['block_close_max_wait_time'] = 1000000
            testFile['block_close_max_events'] = 1000000
            testFile['block_close_max_size'] = 1000000
            testFile['block_close_max_files'] = 1000000

            lumis = []
            for j in range(20):
                lumis.append((i * 20) + j)
            testFile.addRun(Run(143316, *lumis))

            testFile.setAcquisitionEra(acqEra)
            testFile.setProcessingVer("1")
            testFile.setGlobalTag("START54::All")
            testFile.create()
            testFile.setLocation("malpaquet")
            testFile.addParents([parentFiles[i * 2]["lfn"],
                                 parentFiles[i * 2 + 1]["lfn"]])
            testFile.addParents([moreParentFiles[i * 2]["lfn"],
                                 moreParentFiles[i * 2 + 1]["lfn"]])
            childFiles.append(testFile)

        return parentFiles, childFiles

    def verifyData(self, datasetName, files):
        """
        _verifyData_

        Verify the data in DBS3 matches what was uploaded to it.
        """
        (dummy, dummyPrimary, processed, tier) = datasetName.split("/")
        (acqEra, dummyProcVer) = processed.split("-")

        result = self.dbsApi.listDatasets(dataset=datasetName, detail=True,
                                          dataset_access_type="VALID")

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["data_tier_name"], tier,
                         "Error: Wrong data tier.")
        self.assertEqual(result[0]["processing_version"], 1,
                         "Error: Wrong processing version.")
        self.assertEqual(result[0]["acquisition_era_name"], acqEra,
                         "Error: Wrong acquisition era.")
        self.assertEqual(result[0]["physics_group_name"], "NoGroup",
                         "Error: Wrong physics group name.")

        results = self.dbsApi.listBlocks(dataset=datasetName, detail=True)
        self.assertEqual(len(results), len(files) / 5,
                         "Error: Wrong number of blocks.")
        for result in results:
            self.assertEqual(result["block_size"], 5120,
                             "Error: Wrong block size.")
            self.assertEqual(result["file_count"], 5,
                             "Error: Wrong number of files.")
            self.assertEqual(result["open_for_writing"], 0,
                             "Error: Block should be closed.")

        results = self.dbsApi.listFileArray(dataset=datasetName, detail=True)
        for result in results:
            fileObj = None
            for fileObj in files:
                if fileObj["lfn"] == result["logical_file_name"]:
                    break
            else:
                fileObj = None

            self.assertTrue(fileObj is not None, "Error: File not found.")
            self.assertEqual(fileObj["size"], result["file_size"],
                             "Error: File size mismatch.")
            self.assertEqual(fileObj["events"], result["event_count"],
                             "Error: Event count mismatch.")

            dbsParents = self.dbsApi.listFileParents(logical_file_name=fileObj["lfn"])
            if len(dbsParents) == 0:
                self.assertEqual(len(fileObj["parents"]), 0,
                                 "Error: Missing parents.")
            else:
                for dbsParent in dbsParents[0]["parent_logical_file_name"]:
                    fileParent = None
                    for fileParent in fileObj["parents"]:
                        if fileParent["lfn"] == dbsParent:
                            break
                    else:
                        fileParent = None

                    self.assertTrue(fileParent is not None, "Error: Missing parent.")
                self.assertEqual(len(fileObj["parents"]), len(dbsParents[0]["parent_logical_file_name"]),
                                 "Error: Wrong number of parents.")

            runLumis = self.dbsApi.listFileLumiArray(logical_file_name=fileObj["lfn"])
            for runLumi in runLumis:
                fileRun = None
                for fileRun in fileObj["runs"]:
                    if fileRun.run == runLumi["run_num"]:
                        for lumi in runLumi["lumi_section_num"]:
                            self.assertTrue(lumi in fileRun.lumis,
                                            "Error: Missing lumis.")
                        break
                else:
                    fileRun = None

                self.assertTrue(fileRun is not None, "Error: Missing run.")

        self.assertEqual(len(files), len(results), "Error: Files missing.")
        return

    def injectWorkflow(self, workflowName='TestWorkflow',
                       taskPath='/TestWorkflow/ReadingEvents',
                       MaxWaitTime=1,
                       MaxFiles=5,
                       MaxEvents=250000000,
                       MaxSize=99999000000999999):
        """
        _injectWorklow_

        Inject a dummy worklow in DBSBuffer for testing,
        returns the workflow ID
        """
        injectWorkflowDAO = self.bufferFactory("InsertWorkflow")
        workflowID = injectWorkflowDAO.execute(workflowName, taskPath,
                                               MaxWaitTime, MaxFiles,
                                               MaxEvents, MaxSize)
        return workflowID

    @attr("integration")
    def testBasicUpload(self):
        """
        _testBasicUpload_

        Verify that we can successfully upload to DBS3.  Also verify that the
        uploader correctly handles files parentage when uploading.
        """
        self.dbsApi = DbsApi(url=self.dbsUrl)
        config = self.getConfig()
        dbsUploader = DBSUploadPoller(config=config)

        # First test verifies that uploader will poll and then not do anything
        # as the database is empty.
        dbsUploader.algorithm()

        acqEra = "Summer%s" % (int(time.time()))
        parentFiles = self.createParentFiles(acqEra)

        # The algorithm needs to be run twice.  On the first iteration it will
        # create all the blocks and upload one.  On the second iteration it will
        # timeout and upload the second block.
        dbsUploader.algorithm()
        time.sleep(5)
        dbsUploader.algorithm()
        time.sleep(5)

        # Verify the files made it into DBS3.
        self.verifyData(parentFiles[0]["datasetPath"], parentFiles)

        # Inject some more parent files and some child files into DBSBuffer.
        # Run the uploader twice, only the parent files should be added to DBS3.
        (moreParentFiles, childFiles) = \
            self.createFilesWithChildren(parentFiles, acqEra)
        dbsUploader.algorithm()
        time.sleep(5)
        dbsUploader.algorithm()
        time.sleep(5)

        self.verifyData(parentFiles[0]["datasetPath"],
                        parentFiles + moreParentFiles)

        # Run the uploader another two times to upload the child files.  Verify
        # that the child files were uploaded.
        dbsUploader.algorithm()
        time.sleep(5)
        dbsUploader.algorithm()
        time.sleep(5)

        self.verifyData(childFiles[0]["datasetPath"], childFiles)
        return

    @attr("integration")
    def testDualUpload(self):
        """
        _testDualUpload_

        Verify that the dual upload mode works correctly.
        """
        self.dbsApi = DbsApi(url=self.dbsUrl)
        config = self.getConfig()
        dbsUploader = DBSUploadPoller(config=config)
        dbsUtil = DBSBufferUtil()

        # First test verifies that uploader will poll and then not do anything
        # as the database is empty.
        dbsUploader.algorithm()

        acqEra = "Summer%s" % (int(time.time()))
        parentFiles = self.createParentFiles(acqEra)
        (moreParentFiles, childFiles) = \
            self.createFilesWithChildren(parentFiles, acqEra)

        allFiles = parentFiles + moreParentFiles
        allBlocks = []
        for i in range(4):
            DBSBufferDataset(parentFiles[0]["datasetPath"]).create()
            blockName = parentFiles[0]["datasetPath"] + "#" + makeUUID()
            dbsBlock = DBSBufferBlock(blockName,
                                      location="malpaquet",
                                      datasetpath=None)
            dbsBlock.status = "Open"
            dbsBlock.setDataset(parentFiles[0]["datasetPath"], 'data', 'VALID')
            dbsUtil.createBlocks([dbsBlock])
            for fileObj in allFiles[i * 5: (i * 5) + 5]:
                dbsBlock.addFile(fileObj, 'data', 'VALID')
                dbsUtil.setBlockFiles({"block": blockName, "filelfn": fileObj["lfn"]})
                if i < 2:
                    dbsBlock.status = "InDBS"
                dbsUtil.updateBlocks([dbsBlock])
            dbsUtil.updateFileStatus([dbsBlock], "InDBS")
            allBlocks.append(dbsBlock)

        DBSBufferDataset(childFiles[0]["datasetPath"]).create()
        blockName = childFiles[0]["datasetPath"] + "#" + makeUUID()
        dbsBlock = DBSBufferBlock(blockName,
                                  location="malpaquet",
                                  datasetpath=None)
        dbsBlock.status = "InDBS"
        dbsBlock.setDataset(childFiles[0]["datasetPath"], 'data', 'VALID')
        dbsUtil.createBlocks([dbsBlock])
        for fileObj in childFiles:
            dbsBlock.addFile(fileObj, 'data', 'VALID')
            dbsUtil.setBlockFiles({"block": blockName, "filelfn": fileObj["lfn"]})

        dbsUtil.updateFileStatus([dbsBlock], "InDBS")

        dbsUploader.algorithm()
        time.sleep(5)
        dbsUploader.algorithm()
        time.sleep(5)

        self.verifyData(parentFiles[0]["datasetPath"], parentFiles)

        # Change the status of the rest of the parent blocks so we can upload
        # them and the children.
        for dbsBlock in allBlocks:
            dbsBlock.status = "InDBS"
            dbsUtil.updateBlocks([dbsBlock])

        dbsUploader.algorithm()
        time.sleep(5)

        self.verifyData(parentFiles[0]["datasetPath"], parentFiles + moreParentFiles)

        # Run the uploader one more time to upload the children.
        dbsUploader.algorithm()
        time.sleep(5)

        self.verifyData(childFiles[0]["datasetPath"], childFiles)
        return

    def testCloseSettingsPerWorkflow(self):
        """
        _testCloseSettingsPerWorkflow_

        Test the block closing mechanics in the DBS3 uploader,
        this uses a fake dbs api to avoid reliance on external services.
        """
        # Signal trapExit that we are a friend
        os.environ["DONT_TRAP_EXIT"] = "True"
        try:
            # Monkey patch the imports of DbsApi
            from WMComponent.DBS3Buffer import DBSUploadPoller as MockDBSUploadPoller
            MockDBSUploadPoller.DbsApi = MockDbsApi

            # Set the poller and the dbsUtil for verification
            myThread = threading.currentThread()
            (_, dbsFilePath) = mkstemp(dir=self.testDir)
            self.dbsUrl = dbsFilePath
            config = self.getConfig()
            dbsUploader = MockDBSUploadPoller.DBSUploadPoller(config=config)
            dbsUtil = DBSBufferUtil()

            # First test is event based limits and timeout with no new files.
            # Set the files and workflow
            acqEra = "TropicalSeason%s" % (int(time.time()))
            workflowName = 'TestWorkload%s' % (int(time.time()))
            taskPath = '/%s/TestProcessing' % workflowName
            self.injectWorkflow(workflowName, taskPath,
                                MaxWaitTime=2, MaxFiles=100,
                                MaxEvents=150)
            self.createParentFiles(acqEra, nFiles=20,
                                   workflowName=workflowName,
                                   taskPath=taskPath)

            # The algorithm needs to be run twice.  On the first iteration it will
            # create all the blocks and upload one with less than 150 events.
            # On the second iteration the second block is uploaded.
            dbsUploader.algorithm()
            dbsUploader.checkBlocks()
            openBlocks = dbsUtil.findOpenBlocks()
            self.assertEqual(len(openBlocks), 1)
            globalFiles = myThread.dbi.processData("SELECT id FROM dbsbuffer_file WHERE status = 'InDBS'")[0].fetchall()
            notUploadedFiles = myThread.dbi.processData("SELECT id FROM dbsbuffer_file WHERE status = 'NOTUPLOADED'")[
                0].fetchall()
            self.assertEqual(len(globalFiles), 14)
            self.assertEqual(len(notUploadedFiles), 6)
            # Check the fake DBS for data
            fakeDBS = open(self.dbsUrl, 'r')
            fakeDBSInfo = json.load(fakeDBS)
            fakeDBS.close()
            self.assertEqual(len(fakeDBSInfo), 2)
            for block in fakeDBSInfo:
                self.assertTrue('block_events' not in block['block'])
                self.assertEqual(block['block']['file_count'], 7)
                self.assertEqual(block['block']['open_for_writing'], 0)
                self.assertTrue('close_settings' not in block)
            time.sleep(3)
            dbsUploader.algorithm()
            dbsUploader.checkBlocks()
            openBlocks = dbsUtil.findOpenBlocks()
            self.assertEqual(len(openBlocks), 0)
            fakeDBS = open(self.dbsUrl, 'r')
            fakeDBSInfo = json.load(fakeDBS)
            fakeDBS.close()
            self.assertEqual(len(fakeDBSInfo), 3)
            for block in fakeDBSInfo:
                if block['block']['file_count'] != 6:
                    self.assertEqual(block['block']['file_count'], 7)
                self.assertTrue('block_events' not in block['block'])
                self.assertEqual(block['block']['open_for_writing'], 0)
                self.assertTrue('close_settings' not in block)

            # Now check the limit by size and timeout with new files
            acqEra = "TropicalSeason%s" % (int(time.time()))
            workflowName = 'TestWorkload%s' % (int(time.time()))
            taskPath = '/%s/TestProcessing' % workflowName
            self.injectWorkflow(workflowName, taskPath,
                                MaxWaitTime=2, MaxFiles=5,
                                MaxEvents=200000000)
            self.createParentFiles(acqEra, nFiles=16,
                                   workflowName=workflowName,
                                   taskPath=taskPath)
            dbsUploader.algorithm()
            dbsUploader.checkBlocks()
            openBlocks = dbsUtil.findOpenBlocks()
            self.assertEqual(len(openBlocks), 1)
            fakeDBS = open(self.dbsUrl, 'r')
            fakeDBSInfo = json.load(fakeDBS)
            fakeDBS.close()
            self.assertEqual(len(fakeDBSInfo), 6)
            for block in fakeDBSInfo:
                if acqEra in block['block']['block_name']:
                    self.assertEqual(block['block']['file_count'], 5)
                self.assertTrue('block_events' not in block['block'])
                self.assertTrue('close_settings' not in block)
                self.assertEqual(block['block']['open_for_writing'], 0)

            # Put more files, they will go into the same block and then it will be closed
            # after timeout
            time.sleep(3)
            self.createParentFiles(acqEra, nFiles=3,
                                   workflowName=workflowName,
                                   taskPath=taskPath)
            dbsUploader.algorithm()
            dbsUploader.checkBlocks()
            openBlocks = dbsUtil.findOpenBlocks()
            self.assertEqual(len(openBlocks), 0)
            fakeDBS = open(self.dbsUrl, 'r')
            fakeDBSInfo = json.load(fakeDBS)
            fakeDBS.close()
            self.assertEqual(len(fakeDBSInfo), 7)
            for block in fakeDBSInfo:
                if acqEra in block['block']['block_name']:
                    if block['block']['file_count'] < 5:
                        self.assertEqual(block['block']['file_count'], 4)
                    else:
                        self.assertEqual(block['block']['file_count'], 5)
                self.assertTrue('block_events' not in block['block'])
                self.assertEqual(block['block']['open_for_writing'], 0)
                self.assertTrue('close_settings' not in block)

            # Finally test size limits
            acqEra = "TropicalSeason%s" % (int(time.time()))
            workflowName = 'TestWorkload%s' % (int(time.time()))
            taskPath = '/%s/TestProcessing' % workflowName
            self.injectWorkflow(workflowName, taskPath,
                                MaxWaitTime=1, MaxFiles=500,
                                MaxEvents=200000000, MaxSize=2048)
            self.createParentFiles(acqEra, nFiles=7,
                                   workflowName=workflowName,
                                   taskPath=taskPath)
            dbsUploader.algorithm()
            dbsUploader.checkBlocks()
            time.sleep(2)
            dbsUploader.algorithm()
            dbsUploader.checkBlocks()

            self.assertEqual(len(openBlocks), 0)
            fakeDBS = open(self.dbsUrl, 'r')
            fakeDBSInfo = json.load(fakeDBS)
            fakeDBS.close()
            self.assertEqual(len(fakeDBSInfo), 11)
            for block in fakeDBSInfo:
                if acqEra in block['block']['block_name']:
                    if block['block']['file_count'] != 1:
                        self.assertEqual(block['block']['block_size'], 2048)
                        self.assertEqual(block['block']['file_count'], 2)
                self.assertTrue('block_events' not in block['block'])
                self.assertEqual(block['block']['open_for_writing'], 0)
                self.assertTrue('close_settings' not in block)
        except Exception as ex:
            self.fail("We failed at some point in the test: %s" % str(ex))
        finally:
            # We don't trust anyone else with _exit
            del os.environ["DONT_TRAP_EXIT"]
        return

    def testParseDBSException(self):
        """
        test parseDBSException function to parse DBS Go-based server exception message
        """
        exc = """[{"error":{"reason":"DBSError Code:110 Description:DBS DB insert record error Function:dbs.bulkblocks.insertFilesViaChunks Message: Error: concurrency error","message":"7cf3dee6307fdaa1f72d0dc03551fd0e6665f32114d752c37819c238cef7a231 unable to insert files, error DBSError Code:110 Description:DBS DB insert record error Function:dbs.bulkblocks.insertFilesViaChunks Message: Error: concurrency error","function":"dbs.bulkblocks.InsertBulkBlocksConcurrently","code":110},"http":{"method":"POST","code":400,"timestamp":"2022-05-27 19:21:45.67710595 +0000 UTC m=+191328.489143316","path":"/dbs/prod/global/DBSWriter/bulkblocks","user_agent":"DBSClient/Unknown/","x_forwarded_host":"dbs-prod.cern.ch","x_forwarded_for":"137.138.157.32","remote_addr":"137.138.63.204:39950"},"exception":400,"type":"HTTPError","message":"DBSError Code:110 Description:DBS DB insert record error Function:dbs.bulkblocks.InsertBulkBlocksConcurrently Message:7cf3dee6307fdaa1f72d0dc03551fd0e6665f32114d752c37819c238cef7a231 unable to insert files, error DBSError Code:110 Description:DBS DB insert record error Function:dbs.bulkblocks.insertFilesViaChunks Message: Error: concurrency error Error: nested DBSError Code:110 Description:DBS DB insert record error Function:dbs.bulkblocks.insertFilesViaChunks Message: Error: concurrency error"}]"""
        data = parseDBSException(exc)
        expect = 'DBSError Code:110 Description:DBS DB insert record error Function:dbs.bulkblocks.insertFilesViaChunks Message: Error: concurrency error'
        self.assertEqual(data, expect)

        # test another type of exception
        exc = Exception()
        data = parseDBSException(exc)
        self.assertEqual(data, exc)
        self.assertIs(isinstance(exc, Exception), True)


if __name__ == '__main__':
    unittest.main()



"""
ReqMgr request handling.

"""

from builtins import str as newstr, bytes
from future.utils import viewitems, listvalues

import json
import traceback

import cherrypy

import WMCore.ReqMgr.Service.RegExp as rx
from WMCore.Database.CMSCouch import CouchError
from WMCore.Lexicon import sanitizeURL
from WMCore.REST.Format import JSONFormat, PrettyJSONFormat
from WMCore.REST.Server import RESTEntity, restcall, rows
from WMCore.REST.Validation import validate_str
from WMCore.REST.Auth import get_user_info
from WMCore.ReqMgr.DataStructs.AuthzByStatus import AuthzByStatus

from WMCore.ReqMgr.DataStructs.Request import RequestInfo
from WMCore.ReqMgr.DataStructs.ReqMgrConfigDataCache import ReqMgrConfigDataCache
from WMCore.ReqMgr.DataStructs.RequestError import InvalidSpecParameterValue
from WMCore.ReqMgr.DataStructs.RequestStatus import (REQUEST_STATE_LIST, REQUEST_STATE_TRANSITION,
                                                     ACTIVE_STATUS, check_allowed_transition)
from WMCore.ReqMgr.DataStructs.RequestType import REQUEST_TYPES
from WMCore.ReqMgr.Utils.Validation import (validate_request_create_args, validate_request_update_args,
                                            validate_clone_create_args, validateOutputDatasets,
                                            validate_request_priority, workqueue_stat_validation,
                                            isUserAllowed)
from WMCore.Services.RequestDB.RequestDBWriter import RequestDBWriter
from WMCore.Services.WorkQueue.WorkQueue import WorkQueue


class Request(RESTEntity):
    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)
        self.reqmgrAuthzByStatus = AuthzByStatus(config.authz_by_status, config.authorized_roles)
        self.reqmgr_db = api.db_handler.get_db(config.couch_reqmgr_db)
        self.reqmgr_db_service = RequestDBWriter(self.reqmgr_db, couchapp="ReqMgr")
        # this need for the post validtiaon
        self.gq_service = WorkQueue(config.couch_host, config.couch_workqueue_db)

    def _validateGET(self, param, safe):
        # TODO: need proper validation but for now pass everything
        args_length = len(param.args)
        if args_length == 1:
            safe.kwargs["name"] = param.args[0]
            param.args.pop()
            return

        no_multi_key = ["detail", "_nostale", "date_range", "common_dict"]
        for key, value in viewitems(param.kwargs):
            # convert string to list
            if key not in no_multi_key and isinstance(value, (newstr, bytes)):
                param.kwargs[key] = [value]

        detail = param.kwargs.get('detail', True)
        if detail in (False, "false", "False", "FALSE"):
            detail = False

        if "status" in param.kwargs and detail:
            for status in param.kwargs["status"]:
                if status.endswith("-archived"):
                    raise InvalidSpecParameterValue(
                        """Can't retrieve bulk archived status requests with detail option True,
                           set detail=false or use other search arguments""")

        for prop in list(param.kwargs):
            safe.kwargs[prop] = param.kwargs.pop(prop)
        return

    def _validateRequestBase(self, param, safe, valFunc, requestName=None):
        data = cherrypy.request.body.read()
        if data:
            request_args = json.loads(data)
        else:
            request_args = {}

        # first, verify if user is allowed to perform such action
        isUserAllowed(self.reqmgrAuthzByStatus, request_args)

        cherrypy.log('Updating request "%s" with these user-provided args: %s' % (requestName, request_args))

        # In case key args are also passed and request body also exists.
        # If the request.body is dictionary update the key args value as well
        if isinstance(request_args, dict):
            for prop in list(param.kwargs):
                request_args[prop] = param.kwargs.pop(prop)

            if requestName:
                request_args["RequestName"] = requestName
            request_args = [request_args]

        safe.kwargs['workload_pair_list'] = []
        for args in request_args:
            workload, r_args = valFunc(args, self.config, self.reqmgr_db_service, param)
            safe.kwargs['workload_pair_list'].append((workload, r_args))

    def _get_request_names(self, ids):
        "Extract request names from given documents"
        # cherrypy.log("request names %s" % ids)
        doc = {}
        if isinstance(ids, list):
            for rid in ids:
                doc[rid] = 'on'
        elif isinstance(ids, (newstr, bytes)):
            doc[ids] = 'on'

        docs = []
        for key in list(doc):
            if key.startswith('request'):
                rid = key.split('request-')[-1]
                if rid != 'all':
                    docs.append(rid)
                del doc[key]
        return docs

    def _getMultiRequestArgs(self, multiRequestForm):
        request_args = {}
        for prop in multiRequestForm:
            if prop == "ids":
                request_names = self._get_request_names(multiRequestForm["ids"])
            elif prop == "new_status":
                request_args["RequestStatus"] = multiRequestForm[prop]
            # remove this
            # elif prop in ["CustodialSites", "AutoApproveSubscriptionSites"]:
            #    request_args[prop] = [multiRequestForm[prop]]
            else:
                request_args[prop] = multiRequestForm[prop]
        return request_names, request_args

    def _validateMultiRequests(self, param, safe, valFunc):

        data = cherrypy.request.body.read()
        if data:
            request_names, request_args = self._getMultiRequestArgs(json.loads(data))
        else:
            # actually this is error case
            # cherrypy.log(str(param.kwargs))
            request_names, request_args = self._getMultiRequestArgs(param.kwargs)

            for prop in request_args:
                if prop == "RequestStatus":
                    del param.kwargs["new_status"]
                else:
                    del param.kwargs[prop]

            del param.kwargs["ids"]

        isUserAllowed(self.reqmgrAuthzByStatus, request_args)

        safe.kwargs['workload_pair_list'] = []

        for request_name in request_names:
            request_args["RequestName"] = request_name
            workload, r_args = valFunc(request_args, self.config, self.reqmgr_db_service, param)
            safe.kwargs['workload_pair_list'].append((workload, r_args))

        safe.kwargs["multi_update_flag"] = True

    def _getRequestNamesFromBody(self, safe):

        request_names = json.loads(cherrypy.request.body.read())
        safe.kwargs['workload_pair_list'] = request_names
        safe.kwargs["multi_names_flag"] = True

    def validate(self, apiobj, method, api, param, safe):
        # to make validate successful
        # move the validated argument to safe
        # make param empty
        # other wise raise the error
        try:
            if method == 'GET':
                self._validateGET(param, safe)

            elif method == 'PUT':
                args_length = len(param.args)

                if args_length == 1:
                    requestName = param.args[0]
                    param.args.pop()
                else:
                    requestName = None
                self._validateRequestBase(param, safe, validate_request_update_args, requestName)

            elif method == 'POST':
                args_length = len(param.args)
                if args_length == 2 and param.args[0] == "clone":
                    # handles clone workflow.- don't validtate args here
                    param.kwargs['OriginalRequestName'] = param.args[1]
                    param.args.pop()
                    param.args.pop()
                    self._validateRequestBase(param, safe, validate_clone_create_args)
                elif args_length == 1 and param.args[0] == "multi_update":
                    # special case for multi update from browser.
                    param.args.pop()
                    self._validateMultiRequests(param, safe, validate_request_update_args)
                elif args_length == 1 and param.args[0] == "bynames":
                    # special case for multi update from browser.
                    param.args.pop()
                    self._getRequestNamesFromBody(safe)
                else:
                    self._validateRequestBase(param, safe, validate_request_create_args)
        except cherrypy.HTTPError as exc:
            raise exc from None
        except InvalidSpecParameterValue as ex:
            raise ex from None
        except Exception as ex:
            # TODO add proper error message instead of trace back
            msg = traceback.format_exc()
            cherrypy.log("Error: %s" % msg)
            if hasattr(ex, "message"):
                if hasattr(ex.message, '__call__'):
                    msg = ex.message()
                else:
                    msg = str(ex)
            else:
                msg = str(ex)
            raise InvalidSpecParameterValue(msg) from None

    def _maskResult(self, mask, result):
        """
        If a mask of parameters was provided in the query string, then filter
        the request key/values accordingly.
        :param mask: a list of strings (keys of the request dictionary)
        :param result: a dict key'ed by the request name, with the whole
            request dictionary as a value
        :return: updates the result object in place and returns it (dict)
        """

        if len(mask) == 1 and mask[0] == "DAS":
            mask = ReqMgrConfigDataCache.getConfig("DAS_RESULT_FILTER")["filter_list"]

        if len(mask) > 0:
            maskedResult = {}
            for reqName, reqDict in viewitems(result):
                reqInfo = RequestInfo(reqDict)
                maskedResult.setdefault(reqName, {})
                for maskKey in mask:
                    foundValue = reqInfo.get(maskKey, None)
                    maskedResult[reqName].update({maskKey: foundValue})

            return maskedResult
        else:
            return result

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    def get(self, **kwargs):
        """
        Returns request info depending on the conditions set by kwargs
        Currently defined kwargs are following.
        statusList, requestNames, requestType, prepID, inputDataset, outputDataset, dateRange
        If jobInfo is True, returns jobInfomation about the request as well.

        TODO:
        stuff like this has to masked out from result of this call:
            _attachments: {u'spec': {u'stub': True, u'length': 51712, u'revpos': 2, u'content_type': u'application/json'}}
            _id: maxa_RequestString-OVERRIDE-ME_130621_174227_9225
            _rev: 4-c6ceb2737793aaeac3f1cdf591593da4

        """
        ### pop arguments unrelated to the user query
        mask = kwargs.pop("mask", [])
        detail = kwargs.pop("detail", True)
        common_dict = int(kwargs.pop("common_dict", 0))  # modifies the response format
        nostale = kwargs.pop("_nostale", False)

        ### these are the query strings supported by this API
        status = kwargs.get("status", [])
        name = kwargs.get("name", [])
        request_type = kwargs.get("request_type", [])
        prep_id = kwargs.get("prep_id", [])
        inputdataset = kwargs.get("inputdataset", [])
        outputdataset = kwargs.get("outputdataset", [])
        date_range = kwargs.get("date_range", False)
        campaign = kwargs.get("campaign", [])
        team = kwargs.get("team", [])
        mc_pileup = kwargs.get("mc_pileup", [])
        data_pileup = kwargs.get("data_pileup", [])
        requestor = kwargs.get("requestor", [])

        # further tweaks to the couch queries
        if len(status) == 1 and status[0] == "ACTIVE":
            status = ACTIVE_STATUS
        if detail in (False, "false", "False", "FALSE"):
            option = {"include_docs": False}
        else:
            option = {"include_docs": True}
        # everything should be stale view. this only needs for test
        if nostale:
            self.reqmgr_db_service._setNoStale()

        request_info = []
        queryMatched = False  # flag to avoid calling the same view twice
        if len(kwargs) == 2:
            if status and team:
                query_keys = [[t, s] for t in team for s in status]
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("byteamandstatus",
                                                                                 option, query_keys))
                queryMatched = True
            elif status and request_type:
                query_keys = [[s, rt] for rt in request_type for s in status]
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("requestsbystatusandtype",
                                                                                 option, query_keys))
                queryMatched = True
            elif status and requestor:
                query_keys = [[s, r] for r in requestor for s in status]
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("bystatusandrequestor",
                                                                                 option, query_keys))
                queryMatched = True
        elif len(kwargs) == 3:
            if status and request_type and requestor:
                query_keys = [[s, rt, req] for s in status for rt in request_type for req in requestor]
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("bystatusandtypeandrequestor",
                                                                                 option, query_keys))
                queryMatched = True

        # anything else that hasn't matched the query combination above
        if not queryMatched:
            if status:
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("bystatus",
                                                                                 option, status))
            if name:
                request_info.append(self.reqmgr_db_service.getRequestByNames(name))
            if request_type:
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("bytype",
                                                                                 option, request_type))
            if prep_id:
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("byprepid",
                                                                                 option, prep_id))
            if inputdataset:
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("byinputdataset",
                                                                                 option, inputdataset))
            if outputdataset:
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("byoutputdataset",
                                                                                 option, outputdataset))
            if date_range:
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("bydate",
                                                                                 option, date_range))
            if campaign:
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("bycampaign",
                                                                                 option, campaign))
            if mc_pileup:
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("bymcpileup",
                                                                                 option, mc_pileup))
            if data_pileup:
                request_info.append(self.reqmgr_db_service.getRequestByCouchView("bydatapileup",
                                                                                 option, data_pileup))

        # get the intersection of the request data
        result = self._intersection_of_request_info(request_info)

        if not result:
            return []

        result = self._maskResult(mask, result)

        if not option["include_docs"]:
            return list(result)

        # set the return format. default format has request name as a key
        # if is set to one it returns list of dictionary with RequestName field.
        if common_dict == 1:
            response_list = listvalues(result)
        else:
            response_list = [result]
        return rows(response_list)

    def _intersection_of_request_info(self, request_info):
        requests = {}
        if len(request_info) < 1:
            return requests

        request_key_set = set(request_info[0].keys())
        for info in request_info:
            request_key_set = set(request_key_set) & set(info.keys())
        # TODO: need to assume some data maight not contains include docs
        for request_name in request_key_set:
            requests[request_name] = request_info[0][request_name]
        return requests

    def _retrieveResubmissionChildren(self, request_name):
        """
        Fetches all the direct children requests from CouchDB.
        Response from CouchDB view is in the following format:
            [{u'id': u'child_workflow_name',
              u'key': u'parent_workflow_name',
              u'value': 'current_request_status'}]
        :param request_name: string with the parent workflow name
        :return: a list of dictionaries with the parent and child workflow and the child status
        """
        result = self.reqmgr_db.loadView('ReqMgr', 'childresubmissionrequests', keys=[request_name])['rows']
        childrenRequestAndStatus = []
        for childInfo in result:
            childrenRequestAndStatus.append(childInfo)
            childrenRequestAndStatus.extend(self._retrieveResubmissionChildren(childInfo['id']))
        return childrenRequestAndStatus

    def _handleNoStatusUpdate(self, workload, request_args, dn):
        """
        For no-status update, we only support the following parameters:
         1. RequestPriority
         2. Global workqueue statistics, while acquiring a workflow
        """
        if 'RequestPriority' in request_args:
            # Yes, we completely ignore any other arguments posted by the user (web UI case)
            request_args = {'RequestPriority': request_args['RequestPriority']}
            validate_request_priority(request_args)
            # must update three places: GQ elements, workload_cache and workload spec
            self.gq_service.updatePriority(workload.name(), request_args['RequestPriority'])
            report = self.reqmgr_db_service.updateRequestProperty(workload.name(), request_args, dn)
            workload.setPriority(request_args['RequestPriority'])
            workload.saveCouchUrl(workload.specUrl())
            cherrypy.log('Updated priority of "{}" to: {}'.format(workload.name(), request_args['RequestPriority']))
        elif workqueue_stat_validation(request_args):
            report = self.reqmgr_db_service.updateRequestStats(workload.name(), request_args)
            cherrypy.log('Updated workqueue statistics of "{}", with:  {}'.format(workload.name(), request_args))
        else:
            msg = "There are invalid arguments for no-status update: %s" % request_args
            raise InvalidSpecParameterValue(msg)

        return report

    def _handleAssignmentApprovedTransition(self, workload, request_args, dn):
        """
        Allows only two arguments: RequestStatus and RequestPriority
        """
        if "RequestPriority" not in request_args:
            msg = "There are invalid arguments for assignment-approved transition: %s" % request_args
            raise InvalidSpecParameterValue(msg)

        validate_request_priority(request_args)
        report = self.reqmgr_db_service.updateRequestProperty(workload.name(), request_args, dn)
        return report

    def _handleAssignmentStateTransition(self, workload, request_args, dn):
        if ('SoftTimeout' in request_args) and ('GracePeriod' in request_args):
            request_args['HardTimeout'] = request_args['SoftTimeout'] + request_args['GracePeriod']

        # Only allow extra value update for assigned status
        cherrypy.log("Assign request %s, input args: %s ..." % (workload.name(), request_args))
        try:
            workload.updateArguments(request_args)
        except Exception as ex:
            msg = traceback.format_exc()
            cherrypy.log("Error for request args %s: %s" % (request_args, msg))
            raise InvalidSpecParameterValue(str(ex))

        # validate/update OutputDatasets after ProcessingString and AcquisionEra is updated
        request_args['OutputDatasets'] = workload.listOutputDatasets()
        validateOutputDatasets(request_args['OutputDatasets'], workload.getDbsUrl())

        # by default, it contains all unmerged LFNs (used by sites to protect the unmerged area)
        request_args['OutputModulesLFNBases'] = workload.listAllOutputModulesLFNBases()

        # Add parentage relation for step chain, task chain:
        chainMap = workload.getChainParentageSimpleMapping()
        if chainMap:
            request_args["ChainParentageMap"] = chainMap

        # save the spec first before update the reqmgr request status to prevent race condition
        # when workflow is pulled to GQ before site white list is updated
        workload.saveCouch(self.config.couch_host, self.config.couch_reqmgr_db)
        report = self.reqmgr_db_service.updateRequestProperty(workload.name(), request_args, dn)

        return report

    def _handleOnlyStateTransition(self, workload, request_args, dn):
        """
        It handles only the state transition.
        Special handling needed if a request is aborted or force completed.
        """
        # if we got here, then the main workflow has been already validated
        # and the status transition is allowed
        req_status = request_args["RequestStatus"]
        cascade = request_args.get("cascade", False)

        if req_status in ["aborted", "force-complete"]:
            # cancel the workflow first
            self.gq_service.cancelWorkflow(workload.name())

        # cascade option is only supported for these 3 statuses. If set, we need to
        # find all the children requests and perform the same status transition
        if req_status in ["rejected", "closed-out", "announced"] and cascade:
            childrenNamesAndStatus = self._retrieveResubmissionChildren(workload.name())
            msg = "Workflow {} has {} ".format(workload.name(), len(childrenNamesAndStatus))
            msg += "children workflows to have a status transition to: {}".format(req_status)
            cherrypy.log(msg)
            for childInfo in childrenNamesAndStatus:
                if check_allowed_transition(childInfo['value'], req_status):
                    cherrypy.log('Updating request status for {} to {}.'.format(childInfo['id'], req_status))
                    self.reqmgr_db_service.updateRequestStatus(childInfo['id'], req_status, dn)
                else:
                    msg = "Status transition from {} to {} ".format(childInfo['value'], req_status)
                    msg += "not allowed for workflow: {}, skipping it!".format(childInfo['id'])
                    cherrypy.log(msg)
        # then update the original/parent workflow status in couchdb
        cherrypy.log('Updating request status for {} to {}.'.format(workload.name(), req_status))
        report = self.reqmgr_db_service.updateRequestStatus(workload.name(), req_status, dn)
        return report

    def _updateRequest(self, workload, request_args):
        dn = get_user_info().get("dn", "unknown")

        if "RequestStatus" not in request_args:
            report = self._handleNoStatusUpdate(workload, request_args, dn)
        else:
            req_status = request_args["RequestStatus"]
            if len(request_args) == 2 and req_status == "assignment-approved":
                report = self._handleAssignmentApprovedTransition(workload, request_args, dn)
            elif len(request_args) > 1 and req_status == "assigned":
                report = self._handleAssignmentStateTransition(workload, request_args, dn)
            elif len(request_args) == 1 or (len(request_args) == 2 and "cascade" in request_args):
                report = self._handleOnlyStateTransition(workload, request_args, dn)
            else:
                msg = "There are invalid arguments with this status transition: %s" % request_args
                raise InvalidSpecParameterValue(msg)

        if report == 'OK':
            return {workload.name(): "OK"}
        return {workload.name(): "ERROR"}

    @restcall(formats=[('application/json', JSONFormat())])
    def put(self, workload_pair_list):
        """workloadPairList is a list of tuple containing (workload, request_args)"""
        report = []
        for workload, request_args in workload_pair_list:
            result = self._updateRequest(workload, request_args)
            report.append(result)
        return report

    @restcall(formats=[('application/json', JSONFormat())])
    def delete(self, request_name):
        cherrypy.log("INFO: Deleting request document '%s' ..." % request_name)
        try:
            self.reqmgr_db.delete_doc(request_name)
        except CouchError as ex:
            msg = "ERROR: Delete failed."
            cherrypy.log(msg + " Reason: %s" % ex)
            raise cherrypy.HTTPError(404, msg)
            # TODO
        # delete should also happen on WMStats
        cherrypy.log("INFO: Delete '%s' done." % request_name)

    def _update_additional_request_args(self, workload, request_args):
        """
        add to request_args properties which is not initially set from user.
        This data will put in to couchdb.
        Update request_args here if additional information need to be put in couchdb
        """
        request_args['RequestWorkflow'] = sanitizeURL("%s/%s/%s/spec" % (request_args["CouchURL"],
                                                                         request_args["CouchWorkloadDBName"],
                                                                         workload.name()))['url']

        # Add the output datasets if necessary
        # for some bizarre reason OutpuDatasets is list of lists
        request_args['OutputDatasets'] = workload.listOutputDatasets()

        # Add initial priority only for the creation of the request
        request_args['InitialPriority'] = request_args["RequestPriority"]

        return

    @restcall(formats=[('application/json', JSONFormat())])
    def post(self, workload_pair_list, multi_update_flag=False, multi_names_flag=False):
        """
        Create and update couchDB with  a new request.
        request argument is passed from validation
        (validation convert cherrypy.request.body data to argument)

        TODO:
        this method will have some parts factored out so that e.g. clone call
        can share functionality.

        NOTES:
        1) do not strip spaces, #4705 will fails upon injection with spaces;
            currently the chain relies on a number of things coming in #4705
        2) reqInputArgs = Utilities.unidecode(json.loads(body))
            (from ReqMgrRESTModel.putRequest)
        """

        # storing the request document into Couch

        if multi_update_flag:
            return self.put(workload_pair_list)
        if multi_names_flag:
            return self.get(name=workload_pair_list)

        out = []
        for workload, request_args in workload_pair_list:
            self._update_additional_request_args(workload, request_args)

            cherrypy.log("Create request, input args: %s ..." % request_args)
            try:
                workload.saveCouch(request_args["CouchURL"], request_args["CouchWorkloadDBName"],
                                   metadata=request_args)
                out.append({'request': workload.name()})
            except Exception as ex:
                # then it failed to add the spec file as attachment
                # we better delete the original request to avoid confusion in wmstats
                cherrypy.log("Error saving request spec to couch: %s " % str(ex))
                self.delete(request_args['RequestName'])

        return out


class RequestStatus(RESTEntity):
    def __init__(self, app, api, config, mount):
        RESTEntity.__init__(self, app, api, config, mount)

    def validate(self, apiobj, method, api, param, safe):
        validate_str("transition", param, safe, rx.RX_BOOL_FLAG, optional=True)
        args_length = len(param.args)
        if args_length == 1:
            safe.kwargs["transiton"] = param.args[0]
            param.args.pop()
            return

    @restcall(formats=[('application/json', JSONFormat())])
    def get(self, transition):
        """
        Return list of allowed request status.
        If transition, return exhaustive list with all request status
        and their defined transitions.
        """
        if transition == "true":
            return rows([REQUEST_STATE_TRANSITION])
        return rows(REQUEST_STATE_LIST)


class RequestType(RESTEntity):
    def __init__(self, app, api, config, mount):
        RESTEntity.__init__(self, app, api, config, mount)

    def validate(self, apiobj, method, api, param, safe):
        pass

    @restcall(formats=[('application/json', JSONFormat())])
    def get(self):
        return rows(REQUEST_TYPES)



#!/usr/bin/env python

from __future__ import print_function
import os
import unittest
import shutil

from WMCore.WMStats.Config import Config
from WMQuality.REST.RESTBaseUnitTestWithDBBackend import RESTBaseUnitTestWithDBBackend
from WMCore.Lexicon import splitCouchServiceURL

class WMStatsTest(RESTBaseUnitTestWithDBBackend):
    """
    """

    def setUp(self):
        """
        setUP global values
        """
        appport = 19888
        config = TestConfig(appport, os.getenv("COUCHURL"), False);
        self.setConfig(config)
        reqmgrCouchDB = splitCouchServiceURL(config.views.wmstats.reqmgrCouchURL)[1]
        wmstatsCouchDB = splitCouchServiceURL(config.views.wmstats.wmstatsCouchURL)[1]
        self.setCouchDBs([(reqmgrCouchDB, "ReqMgr"), (wmstatsCouchDB, "WMStats")])
        self.setSchemaModules([])
        RESTBaseUnitTestWithDBBackend.setUp(self)

    def tearDown(self):
        RESTBaseUnitTestWithDBBackend.tearDown(self)

    def testRequest(self):
        # add request related REST API test
        #
        #self.jsonSender.put('request/' + schema['RequestName'], schema)
        #print self.jsonSender.get('request', incoming_headers=self.adminHeader)
        print(self.jsonSender.get('requests'))

if __name__ == '__main__':

    unittest.main()



#!/usr/bin/env python
"""
_IDFromFilesetWorkflow_

Oracle implementation of Subscriptions.IDFromFilesetWorkflow
"""




from WMCore.WMBS.MySQL.Subscriptions.IDFromFilesetWorkflow import IDFromFilesetWorkflow as IDFromFilesetWorkflowMySQL

class IDFromFilesetWorkflow(IDFromFilesetWorkflowMySQL):
    pass



#!/usr/bin/env python
"""
_SiblingSubscriptionsFailed_

MySQL implementation of Subscription.SiblingSubscriptionsFailed
"""

from WMCore.Database.DBFormatter import DBFormatter

class SiblingSubscriptionsFailed(DBFormatter):
    # Find files that have failed processing and mark them as complete for the
    # cleanup subscription.  These files should not be cleaned up as they may be
    # used to investigate why the job failed.  This should only be run after the
    # input fileset has been closed.
    sql = """SELECT DISTINCT wmbs_fileset_files.fileid AS fileid,
                             :subscription AS subscription FROM wmbs_fileset_files
               INNER JOIN wmbs_subscription ON
                 wmbs_fileset_files.fileset = wmbs_subscription.fileset
               INNER JOIN wmbs_sub_files_failed ON
                 wmbs_fileset_files.fileid = wmbs_sub_files_failed.fileid AND
                 wmbs_subscription.id = wmbs_sub_files_failed.subscription
             WHERE wmbs_fileset_files.fileset = :fileset"""

    delete = """DELETE FROM wmbs_sub_files_available
                WHERE subscription = :subscription AND
                      fileid = :fileid"""

    insert = """INSERT IGNORE INTO wmbs_sub_files_complete (subscription, fileid) VALUES
                  (:subscription, :fileid)"""

    def execute(self, subscription, fileset, conn = None, transaction = False):
        results = self.dbi.processData(self.sql, {"fileset": fileset,
                                                  "subscription": subscription},
                                       conn = conn, transaction = transaction)
        failedFiles = self.formatDict(results)

        if len(failedFiles) > 0:
            self.dbi.processData(self.delete, failedFiles, conn = conn,
                                 transaction = transaction)
            self.dbi.processData(self.insert, failedFiles, conn = conn,
                                 transaction = transaction)
        return



#!/usr/bin/env python
# encoding: utf-8
"""
Locale_t.py

Created by Dave Evans on 2011-05-23.
Copyright (c) 2011 Fermilab. All rights reserved.
"""

import unittest
from WMCore.Locale import Locale
from WMCore.Configuration import Configuration

class Locale_t(unittest.TestCase):

    def testA(self):
        """check sections/fields exist in Locale template"""
        sections = ['reqmgr', 'gwq', 'lwq', 'agent', 'crabserver', 'couch', 'mysql']
        for s in sections:
            self.assertTrue(hasattr(Locale.locale, s))

    def testB(self):
        """check set/get"""

        Locale.locale.mysql.url = "mysql://whatever:3306"
        Locale.locale.couch.port = "1999"

        self.assertEqual(Locale.locale.mysql.url, "mysql://whatever:3306")
        self.assertEqual(Locale.locale.couch.port, "1999")

    def testC(self):
        """test addition of locale to configuration instance"""

        conf = Configuration()

        conf + Locale
        self.assertTrue(hasattr(conf, 'locale'))



if __name__ == '__main__':
    unittest.main()



"""
_WMConfigCache_t_

Test class for the WMConfigCache
"""
from __future__ import print_function, division

import unittest
import time

from Utils.PythonVersion import PY3

from WMCore.Cache.GenericDataCache import GenericDataCache, CacheExistException, \
                          CacheWithWrongStructException, MemoryCacheStruct

from Utils.PythonVersion import PY3

class Foo(object):
    pass


class GenericDataCacheTest(unittest.TestCase):

    def setUp(self):
        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def testBasic(self):
        """
        _testBasic_

        Basic stuff.
        """
        mc = MemoryCacheStruct(1, lambda x: int(time.time()), kwargs={'x':1})
        self.assertIsNone(mc.data)
        self.assertEqual(mc.lastUpdated, -1)

        GenericDataCache.registerCache("test", mc)
        with self.assertRaises(CacheExistException):
            GenericDataCache.registerCache("test", mc)
        with self.assertRaises(CacheWithWrongStructException):
            GenericDataCache.registerCache("test2", {'a': 1})
        mc2 = GenericDataCache.getCacheData('test')
        before = mc2.getData()
        time.sleep(2)
        after = mc2.getData()
        self.assertFalse(before == after)
        self.assertFalse(mc2.lastUpdated == -1)

        return

    def testBasicInit(self):
        """
        _testBasicInit_

        Test init values
        """
        mc = MemoryCacheStruct(0, lambda x: x, initCacheValue=Foo())
        self.assertIsInstance(mc.data, Foo)

        mc1 = MemoryCacheStruct(0, lambda x: sum(x), [], kwargs={'x': [1, 2]})
        self.assertEqual(mc1.data, [])
        after = mc1.getData()
        self.assertEqual(after, 3)

        mc2 = MemoryCacheStruct(0, lambda x: x, {}, kwargs={'x': {'one':1, 'two':2}})
        self.assertEqual(mc2.data, {})
        after = mc2.getData()
        self.assertCountEqual(after.keys(), ['one', 'two']) if PY3 else self.assertItemsEqual(after.keys(), ['one', 'two'])

        return

    def testExists(self):
        """
        Tests whether a given cache already exists
        """
        mc = MemoryCacheStruct(1, lambda x: int(time.time()), kwargs={'x':1})
        self.assertFalse(GenericDataCache.cacheExists("tCache"))

        GenericDataCache.registerCache("tCache", mc)
        self.assertTrue(GenericDataCache.cacheExists("tCache"))
        self.assertFalse(GenericDataCache.cacheExists("tCache2"))

if __name__ == "__main__":
    unittest.main()



"""
Unit tests for Unified/Monitor.py module

Author: Valentin Kuznetsov <vkuznet [AT] gmail [DOT] com>
"""
from __future__ import division, print_function

from future.utils import viewitems

import time
# system modules
import unittest
from copy import deepcopy

# WMCore modules
from WMCore.MicroService.MSMonitor.MSMonitor import MSMonitor
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.Emulators.ReqMgrAux.MockReqMgrAux import MockReqMgrAux


class MSMonitorTest(EmulatedUnitTestCase):
    "Unit test for Monitor module"

    def setUp(self):
        "init test class"
        self.msConfig = {'verbose': False,
                         'group': 'DataOps',
                         'interval': 1 * 60,
                         'updateInterval': 0,
                         'enableStatusTransition': True,
                         'reqmgr2Url': 'https://cmsweb-testbed.cern.ch/reqmgr2',
                         'reqmgrCacheUrl': 'https://cmsweb-testbed.cern.ch/couchdb/reqmgr_workload_cache',
                         'phedexUrl': 'https://cmsweb-testbed.cern.ch/phedex/datasvc/json/prod',
                         'dbsUrl': 'https://cmsweb-testbed.cern.ch/dbs/int/global/DBSReader',
                         'rucioAccount': "wma_test",
                         'rucioUrl': "http://cms-rucio-int.cern.ch",
                         'rucioAuthUrl': "https://cms-rucio-auth-int.cern.ch"}

        self.ms = MSMonitor(self.msConfig)
        self.ms.reqmgrAux = MockReqMgrAux()
        super(MSMonitorTest, self).setUp()

    def testUpdateCaches(self):
        """
        Test the getCampaignConfig method
        """
        campaigns, transfersDocs = self.ms.updateCaches()
        self.assertNotEqual(transfersDocs, [])
        self.assertEqual(len(transfersDocs[0]['transfers']), 1)
        self.assertTrue(time.time() > transfersDocs[0]['lastUpdate'], 1)

        self.assertNotEqual(campaigns, [])
        for cname, cdict in viewitems(campaigns):
            self.assertEqual(cname, cdict['CampaignName'])
            self.assertEqual(isinstance(cdict, dict), True)
            self.assertNotEqual(cdict.get('CampaignName', {}), {})

    def testGetTransferInfo(self):
        """
        Test the getTransferInfo method
        """
        _, transfersDocs = self.ms.updateCaches()
        transfersDocs[0]['transfers'] = []
        originalTransfers = deepcopy(transfersDocs)
        self.ms.getTransferInfo(transfersDocs)

        self.assertNotEqual(transfersDocs, [])
        self.assertEqual(len(transfersDocs), len(originalTransfers))
        for rec in transfersDocs:
            self.assertEqual(isinstance(rec, dict), True)
            keys = sorted(['workflowName', 'lastUpdate', 'transfers'])
            self.assertEqual(keys, sorted(rec.keys()))
            self.assertTrue(time.time() >= rec['lastUpdate'])

    def testCompletion(self):
        """
        Test the completion method
        """
        campaigns, transfersDocs = self.ms.updateCaches()
        transfersDocs.append(deepcopy(transfersDocs[0]))
        transfersDocs.append(deepcopy(transfersDocs[0]))
        transfersDocs[0]['transfers'] = []
        transfersDocs[0]['workflowName'] = 'workflow_0'
        transfersDocs[1]['transfers'][0]['completion'].append(100)
        transfersDocs[1]['workflowName'] = 'workflow_1'
        transfersDocs[2]['workflowName'] = 'workflow_2'
        self.assertEqual(len(transfersDocs), 3)

        completedWfs = self.ms.getCompletedWorkflows(transfersDocs, campaigns)
        self.assertEqual(len(completedWfs), 2)

    def testUpdateTransferInfo(self):
        """
        Test the updateTransferInfo method
        """
        _, transferRecords = self.ms.updateCaches()
        failed = self.ms.updateTransferDocs(transferRecords, workflowsToSkip=[])
        self.assertEqual(len(failed), len(transferRecords))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
MakeCRICMockFiles
Program to create mock SiteDB JSON files used by the SiteDB mock-based emulator
"""

from __future__ import division, print_function
from future import standard_library
standard_library.install_aliases()

import json
import os
from urllib.error import HTTPError

from WMCore.Services.CRIC.CRIC import CRIC
from WMCore.WMBase import getTestBase

if __name__ == '__main__':
    calls = ['people', 'site-names', 'data-processing']

    additionals = {
        'site-names': [
            {"site_name": "T2_XX_SiteA", "type": "psn", "alias": "T2_XX_SiteA"},
            {"site_name": "T2_XX_SiteB", "type": "psn", "alias": "T2_XX_SiteB"},
            {"site_name": "T2_XX_SiteC", "type": "psn", "alias": "T2_XX_SiteC"},
            {"site_name": "T2_XX_SiteA", "type": "phedex", "alias": "T2_XX_SiteA"},
            {"site_name": "T2_XX_SiteB", "type": "phedex", "alias": "T2_XX_SiteB"},
            {"site_name": "T2_XX_SiteC", "type": "phedex", "alias": "T2_XX_SiteC"},
        ],
        'data-processing': [
            {"phedex_name": "T2_XX_SiteA", "psn_name": "T2_XX_SiteA"},
            {"phedex_name": "T2_XX_SiteB", "psn_name": "T2_XX_SiteB"},
            {"phedex_name": "T2_XX_SiteC", "psn_name": "T2_XX_SiteC"},
        ]
    }

    lookup = {}

    outFile = 'CRICMockData.json'
    outFilename = os.path.join(getTestBase(), '..', 'data', 'Mock', outFile)
    print("Creating the file %s with mocked CRIC data" % outFilename)

    cric = CRIC()
    for callname in calls:
        print("Querying %s ..." % callname)
        try:
            if callname == 'people':
                result = cric._CRICUserQuery(callname)
            else:
                result = cric._CRICSiteQuery(callname)
        except HTTPError as exc:
            result = 'Raises HTTPError'

        if callname in ["site-names", "data-processing"]:
            result.extend(additionals[callname])
        else:
            result = result
        lookup.update({callname:result})

    with open(outFilename, 'w') as mockData:
        json.dump(lookup, mockData, indent=1, separators=(',', ': '), sort_keys=True)



#!/usr/bin/env python
"""
_Tier1ReRecoWorkload_
"""

from WMCore.WMSpec.WMWorkload import newWorkload

#  //
# // Set up the basic workload task and step structure
#//
def createWorkload(name="BasicProcessing"):
    workload = newWorkload(name)
    workload.setOwner("DMWMTest")
    workload.setStartPolicy('DatasetBlock')
    workload.setEndPolicy('SingleShot')
    workload.setCampaign("TestCampaign")
    workload.setPrepID("TestPrepID")
    #  //
    # // set up the production task
    #//
    rereco = workload.newTask("ReReco1")
    rereco.setTaskType("Processing")
    rerecoCmssw = rereco.makeStep("cmsRun1")
    rerecoCmssw.setStepType("CMSSW")
    skimStageOut = rerecoCmssw.addStep("stageOut1")
    skimStageOut.setStepType("StageOut")
    skimLogArch = rerecoCmssw.addStep("logArch1")
    skimLogArch.setStepType("LogArchive")
    rereco.applyTemplates()
    rereco.setSplittingAlgorithm("FileBased", files_per_job=1)
    rereco.addInputDataset(
        primary = "Cosmics",
        processed = "CRAFT09-PromptReco-v1",
        tier = "RECO",
        dbsurl = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader")

    #  //
    # // rereco cmssw step
    #//
    #
    # TODO: Anywhere helper.data is accessed means we need a method added to the
    # type based helper class to provide a clear API.
    rerecoCmsswHelper = rerecoCmssw.getTypeHelper()


    rerecoCmsswHelper.cmsswSetup(
        "CMSSW_3_1_2",
        softwareEnvironment=" . /uscmst1/prod/sw/cms/bashrc prod"
        )

    rerecoCmsswHelper.setDataProcessingConfig(
        "cosmics", "promptReco", globalTag="GLOBAL::BALLS",
        writeTiers=['RECO'])

    rerecoCmsswHelper.addOutputModule(
        "outputRECO", primaryDataset="Primary",
        processedDataset="Processed",
        dataTier="RECO")

    #Add a stageOut step
    skimStageOutHelper = skimStageOut.getTypeHelper()
    skimLogArchHelper  = skimLogArch.getTypeHelper()


    rereco.addGenerator("BasicNaming")
    return workload



#!/usr/bin/env python
"""
_DeleteJobs_

MySQL implementation for creating a deleting a job
"""


from WMCore.Database.DBFormatter import DBFormatter

class DeleteJobs(DBFormatter):
    """
    _DeleteJobs_

    Delete jobs from bl_runjob
    """


    sql = """DELETE FROM bl_runjob WHERE id = :id
    """
    def execute(self, jobs, conn = None, transaction = False):
        """
        _execute_

        Delete jobs
        Expects a list of IDs
        """

        if len(jobs) == 0:
            return

        binds = []
        for jobid in jobs:
            binds.append({'id': jobid})

        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)

        return



#!/usr/bin/env python
"""
_GetChecksum_

MySQL implementation of DBSBufferFiles.GetChecksum
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetChecksum(DBFormatter):
    sql = """SELECT cst.type AS cktype, fcs.cksum AS cksum FROM
               dbsbuffer_file_checksums fcs INNER JOIN
               dbsbuffer_checksum_type cst
               ON fcs.typeid = cst.id
               WHERE fcs.fileid = :fileid"""

    def formatResult(self, result):
        """
        _formatResult_

        Turn the query results into a dictionary that is keyed by
        checksum type.
        """
        formattedResult = {}

        dictVersion = DBFormatter.formatDict(self, result)

        for resultRow in dictVersion:
            formattedResult[resultRow["cktype"]] = resultRow["cksum"]

        return formattedResult

    def execute(self, fileid, conn = None, transaction = False):
        """
        _execute_

        Retrieve checksum information for a file.  The result is returned in
        the form of a dict that is keyed by checksum type.
        """
        result = self.dbi.processData(self.sql, {"fileid": fileid}, conn = conn,
                                      transaction = transaction)
        return self.formatResult(result)



#!/bin/env python


"""
_ExponentialAlgo_

It keeps getting bigger and bigger
"""

from math import pow

from WMComponent.RetryManager.PlugIns.RetryAlgoBase import RetryAlgoBase


class ExponentialAlgo(RetryAlgoBase):
    """
    _ExponentialAlgo_

    Delay more each retry by some exponential factor
    """


    def isReady(self, job, cooloffType):
        """
        Actual function that does the work

        """

        # Get the cooloff time
        baseTimeoutDict = self.getAlgoParam(job['jobType'])
        baseTimeout = baseTimeoutDict.get(cooloffType.lower(), 10)
        cooloffTime = pow(baseTimeout, job['retry_count'])

        currentTime = self.timestamp()
        if currentTime - job['state_time'] > cooloffTime:
            return True
        else:
            return False
